# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from __future__ import division

__all__ = ['GWP', 'ODP', 'logP', 'GWP_methods', 'GWP_data', 'ODP_methods', 
           'ODP_data', 'logP_methods', 'CRClogPDict', 'SyrresDict2']

import os
import pandas as pd


folder = os.path.join(os.path.dirname(__file__), 'Environment')


### Global Warming Potentials

GWP_data = pd.read_csv(os.path.join(folder,
                       'Official Global Warming Potentials.tsv'), sep='\t',
                       index_col=0)

ODP_data = pd.read_csv(os.path.join(folder,
                       'Ozone Depletion Potentials.tsv'), sep='\t',
                       index_col=0)

CRClogPDict = pd.read_csv(os.path.join(folder,
                       'CRC logP table.tsv'), sep='\t',
                       index_col=0)

SyrresDict2 = pd.read_csv(os.path.join(folder,
                       'Syrres logP data.csv.gz'), sep='\t',
                       index_col=0, compression='gzip')

IPCC100 = 'IPCC (2007) 100yr'
IPCC100SAR = 'IPCC (2007) 100yr-SAR'
IPCC20 = 'IPCC (2007) 20yr'
IPCC500 = 'IPCC (2007) 500yr'
NONE = 'None'
GWP_methods = [IPCC100, IPCC100SAR, IPCC20, IPCC500]


def GWP(CASRN, AvailableMethods=False, Method=None):
    r'''This function handles the retrieval of a chemical's Global Warming
    Potential, relative to CO2. Lookup is based on CASRNs. Will automatically
    select a data source to use if no Method is provided; returns None if the
    data is not available.

    Returns the GWP for the 100yr outlook by default.

    Parameters
    ----------
    CASRN : string
        CASRN [-]

    Returns
    -------
    GWP : float
        Global warming potential, [(impact/mass chemical)/(impact/mass CO2)]
    methods : list, only returned if AvailableMethods == True
        List of methods which can be used to obtain GWP with the
        given inputs

    Other Parameters
    ----------------
    Method : string, optional
        The method name to use. Accepted methods are IPCC (2007) 100yr',
        'IPCC (2007) 100yr-SAR', 'IPCC (2007) 20yr', and 'IPCC (2007) 500yr'. 
        All valid values are also held in the list GWP_methods.
    AvailableMethods : bool, optional
        If True, function will determine which methods can be used to obtain
        the GWP for the desired chemical, and will return methods
        instead of the GWP

    Notes
    -----
    All data is from [1]_, the official source. Several chemicals are available
    in [1]_ are not included here as they do not have a CAS.
    Methods are 'IPCC (2007) 100yr', 'IPCC (2007) 100yr-SAR',
    'IPCC (2007) 20yr', and 'IPCC (2007) 500yr'.

    Examples
    --------
    Methane, 100-yr outlook

    >>> GWP(CASRN='74-82-8')
    25.0

    References
    ----------
    .. [1] IPCC. "2.10.2 Direct Global Warming Potentials - AR4 WGI Chapter 2:
       Changes in Atmospheric Constituents and in Radiative Forcing." 2007.
       https://www.ipcc.ch/publications_and_data/ar4/wg1/en/ch2s2-10-2.html.
    '''
    def list_methods():
        methods = []
        if CASRN in GWP_data.index:
            methods.append(IPCC100)
            if not pd.isnull(GWP_data.at[CASRN, 'SAR 100yr']):
                methods.append(IPCC100SAR)
            methods.append(IPCC20)
            methods.append(IPCC500)
        methods.append(NONE)
        return methods
    if AvailableMethods:
        return list_methods()
    if not Method:
        Method = list_methods()[0]

    if Method == IPCC100:
        return float(GWP_data.at[CASRN, '100yr GWP'])
    elif Method == IPCC100SAR:
        return float(GWP_data.at[CASRN, 'SAR 100yr'])
    elif Method == IPCC20:
        return float(GWP_data.at[CASRN, '20yr GWP'])
    elif Method == IPCC500:
        return float(GWP_data.at[CASRN, '500yr GWP'])
    elif Method == NONE:
        return None
    else:
        raise Exception('Failure in in function')


### Ozone Depletion Potentials


ODP2MAX = 'ODP2 Max'
ODP2MIN = 'ODP2 Min'
ODP2STR = 'ODP2 string'
ODP2LOG = 'ODP2 logarithmic average'
ODP1MAX = 'ODP1 Max'
ODP1MIN = 'ODP1 Min'
ODP1STR = 'ODP1 string'
ODP1LOG = 'ODP1 logarithmic average'
NONE = 'None'
ODP_methods = [ODP2MAX, ODP1MAX, ODP2LOG, ODP1LOG, ODP2MIN, ODP1MIN, ODP2STR, ODP1STR]


def ODP(CASRN, AvailableMethods=False, Method=None):
    r'''This function handles the retrieval of a chemical's Ozone Depletion
    Potential, relative to CFC-11 (trichlorofluoromethane). Lookup is based on
    CASRNs. Will automatically select a data source to use if no Method is
    provided; returns None if the data is not available.

    Returns the ODP of a chemical according to [2]_ when a method is not
    specified. If a range is provided in [2]_, the highest value is returned.

    Parameters
    ----------
    CASRN : string
        CASRN [-]

    Returns
    -------
    ODP : float or str
        Ozone Depletion potential, [(impact/mass chemical)/(impact/mass CFC-11)];
        if method selected has `string` in it, this will be returned as a
        string regardless of if a range is given or a number
    methods : list, only returned if AvailableMethods == True
        List of methods which can be used to obtain ODP with the
        given inputs

    Other Parameters
    ----------------
    Method : string, optional
        The method name to use. Accepted methods are 'ODP2 Max', 'ODP2 Min', 
        'ODP2 string', 'ODP2 logarithmic average', and methods for older values
        are 'ODP1 Max', 'ODP1 Min', 'ODP1 string', and 'ODP1 logarithmic average'.
        All valid values are also held in the list ODP_methods.
    Method : string, optional
        A string for the method name to use, as defined by constants in
        ODP_methods
    AvailableMethods : bool, optional
        If True, function will determine which methods can be used to obtain
        the ODP for the desired chemical, and will return methods
        instead of the ODP

    Notes
    -----
    Values are tabulated only for a small number of halogenated hydrocarbons,
    responsible for the largest impact. The original values of ODP as defined
    in the Montreal Protocol are also available, as methods with the `ODP1`
    prefix.

    All values are somewhat emperical, as actual reaction rates of chemicals
    with ozone depend on temperature which depends on latitude, longitude,
    time of day, weather, and the concentrations of other pollutants.

    All data is from [1]_. Several mixtures listed in [1]_ are not included
    here as they are not pure species.
    Methods for values in [2]_ are 'ODP2 Max', 'ODP2 Min', 'ODP2 string',
    'ODP2 logarithmic average',  and methods for older values are 'ODP1 Max',
    'ODP1 Min', 'ODP1 string', and 'ODP1 logarithmic average'.

    Examples
    --------
    Dichlorotetrafluoroethane, according to [2]_.

    >>> ODP(CASRN='76-14-2')
    0.58

    References
    ----------
    .. [1] US EPA, OAR. "Ozone-Depleting Substances." Accessed April 26, 2016.
       https://www.epa.gov/ozone-layer-protection/ozone-depleting-substances.
    .. [2] WMO (World Meteorological Organization), 2011: Scientific Assessment
       of Ozone Depletion: 2010. Global Ozone Research and Monitoring
       Project-Report No. 52, Geneva, Switzerland, 516 p.
       https://www.wmo.int/pages/prog/arep/gaw/ozone_2010/documents/Ozone-Assessment-2010-complete.pdf
    '''
    def list_methods():
        methods = []
        if CASRN in ODP_data.index:
            if not pd.isnull(ODP_data.at[CASRN, 'ODP2 Max']):
                methods.append(ODP2MAX)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP1 Max']):
                methods.append(ODP1MAX)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP2 Design']):
                methods.append(ODP2LOG)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP1 Design']):
                methods.append(ODP1LOG)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP2 Min']):
                methods.append(ODP2MIN)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP1 Min']):
                methods.append(ODP1MIN)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP2']):
                methods.append(ODP2STR)
            if not pd.isnull(ODP_data.at[CASRN, 'ODP1']):
                methods.append(ODP1STR)
        methods.append(NONE)
        return methods
    if AvailableMethods:
        return list_methods()
    if not Method:
        Method = list_methods()[0]

    if Method == ODP2MAX:
        return float(ODP_data.at[CASRN, 'ODP2 Max'])
    elif Method == ODP1MAX:
        return float(ODP_data.at[CASRN, 'ODP1 Max'])
    elif Method == ODP2MIN:
        return float(ODP_data.at[CASRN, 'ODP2 Min'])
    elif Method == ODP1MIN:
        return float(ODP_data.at[CASRN, 'ODP1 Min'])
    elif Method == ODP2LOG:
        return float(ODP_data.at[CASRN, 'ODP2 Design'])
    elif Method == ODP1LOG:
        return float(ODP_data.at[CASRN, 'ODP1 Design'])
    elif Method == ODP2STR:
        return str(ODP_data.at[CASRN, 'ODP2'])
    elif Method == ODP1STR:
        return str(ODP_data.at[CASRN, 'ODP1'])
    elif Method == NONE:
        return None
    else:
        raise Exception('Failure in in function')


### log P


SYRRES = 'SYRRES'
CRC = 'CRC'
NONE = 'NONE'
logP_methods = [SYRRES, CRC]


def logP(CASRN, AvailableMethods=False, Method=None):
    r'''This function handles the retrieval of a chemical's octanol-water
    partition coefficient. Lookup is based on CASRNs. Will automatically
    select a data source to use if no Method is provided; returns None if the
    data is not available.

    Parameters
    ----------
    CASRN : string
        CASRN [-]

    Returns
    -------
    logP : float
        Octanol-water partition coefficient, [-]
    methods : list, only returned if AvailableMethods == True
        List of methods which can be used to obtain logP with the
        given inputs

    Other Parameters
    ----------------
    Method : string, optional
        The method name to use. Accepted methods are 'SYRRES', or 'CRC', 
        All valid values are also held in the list logP_methods.
    AvailableMethods : bool, optional
        If True, function will determine which methods can be used to obtain
        the logP for the desired chemical, and will return methods
        instead of the logP

    Notes
    -----
    .. math::
        \log P_{ oct/wat} = \log\left(\frac{\left[{solute}
        \right]_{ octanol}^{un-ionized}}{\left[{solute}
        \right]_{ water}^{ un-ionized}}\right)

    Examples
    --------
    >>> logP('67-56-1')
    -0.74

    References
    ----------
    .. [1] Syrres. 2006. KOWWIN Data, SrcKowData2.zip.
       http://esc.syrres.com/interkow/Download/SrcKowData2.zip
    .. [2] Haynes, W.M., Thomas J. Bruno, and David R. Lide. CRC Handbook of
       Chemistry and Physics, 95E. Boca Raton, FL: CRC press, 2014.
    '''
    def list_methods():
        methods = []
        if CASRN in CRClogPDict.index:
            methods.append(CRC)
        if CASRN in SyrresDict2.index:
            methods.append(SYRRES)
        methods.append(NONE)
        return methods
    if AvailableMethods:
        return list_methods()
    if not Method:
        Method = list_methods()[0]

    if Method == CRC:
        return float(CRClogPDict.at[CASRN, 'logP'])
    elif Method == SYRRES:
        return float(SyrresDict2.at[CASRN, 'logP'])
    elif Method == NONE:
        return None
    else:
        raise Exception('Failure in in function')



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from numpy.testing import assert_allclose
import pytest
import pandas as pd
from thermo.critical import *

def test_data_IUPAC():
    MW_sum = _crit_IUPAC['MW'].sum()
    assert_allclose(MW_sum,122998.43799999992)

    Tc_sum = _crit_IUPAC['Tc'].sum()
    assert_allclose(Tc_sum, 462157.51300000004)

    Pc_sum = _crit_IUPAC['Pc'].sum()
    assert_allclose(Pc_sum, 2063753000.0)

    Vc_sum = _crit_IUPAC['Vc'].sum()
    assert_allclose(Vc_sum, 0.19953190000000001)

    Zc_sum = _crit_IUPAC['Zc'].sum()
    assert_allclose(Zc_sum, 109.892)

    assert _crit_IUPAC.shape == (810, 7)
    assert _crit_IUPAC.index.is_unique


def test_data_Matthews():
    MW_sum = _crit_Matthews['MW'].sum()
    assert_allclose(MW_sum, 19541.760399999999)

    Tc_sum = _crit_Matthews['Tc'].sum()
    assert_allclose(Tc_sum, 65343.210900000005)

    Pc_sum = _crit_Matthews['Pc'].sum()
    assert_allclose(Pc_sum, 579365204.25)

    Vc_sum = _crit_Matthews['Vc'].sum()
    assert_allclose(Vc_sum, 0.014921000000000002)

    Zc_sum = _crit_Matthews['Zc'].sum()
    assert_allclose(Zc_sum, 12.141000000000002)

    assert _crit_Matthews.shape == (120, 6)
    assert _crit_Matthews.index.is_unique


def test_data_CRC():
    Tc_sum = _crit_CRC['Tc'].sum()
    assert_allclose(Tc_sum, 514092.75)

    Pc_sum = _crit_CRC['Pc'].sum()
    assert_allclose(Pc_sum, 2700259000.0)

    Vc_sum = _crit_CRC['Vc'].sum()
    assert_allclose(Vc_sum, 0.38876929999999998)

    Zc_sum = _crit_CRC['Zc'].sum()
    assert_allclose(Zc_sum, 207.98663028416496, 1e-6)

    assert _crit_CRC.shape == (861, 8)
    assert _crit_CRC.index.is_unique

    Tc_error_sum = _crit_CRC['Tc_error'].sum()
    assert_allclose(Tc_error_sum, 2444.24)

    Pc_error_sum = _crit_CRC['Pc_error'].sum()
    assert_allclose(Pc_error_sum, 1.2587e+08)

    Vc_error_sum = _crit_CRC['Vc_error'].sum()
    assert_allclose(Vc_error_sum, 0.014151)


def test_data_PSRKR4():
    Tc_sum = _crit_PSRKR4['Tc'].sum()
    assert_allclose(Tc_sum, 597984.0)

    Pc_sum = _crit_PSRKR4['Pc'].sum()
    assert_allclose(Pc_sum, 3708990509)

    Vc_sum = _crit_PSRKR4['Vc'].sum()
    assert_allclose(Vc_sum, 0.40726849999999998)

    Zc_sum = _crit_PSRKR4['Zc'].sum()
    assert_allclose(Zc_sum, 251.29839643655527, 1e-6)

    omega_sum = _crit_PSRKR4['omega'].sum()
    assert_allclose(omega_sum, 410.50560000000002)

    assert _crit_PSRKR4.shape == (995, 6)
    assert _crit_PSRKR4.index.is_unique


def test_data_PassutDanner():
    Tc_sum = _crit_PassutDanner['Tc'].sum()
    assert_allclose(Tc_sum, 111665.28333333334)

    Pc_sum = _crit_PassutDanner['Pc'].sum()
    assert_allclose(Pc_sum, 579756767.55527318)

    omega_sum = _crit_PassutDanner['omega'].sum()
    assert_allclose(omega_sum, 65.567000000000007)

    assert _crit_PassutDanner.shape == (192, 4)
    assert _crit_PassutDanner.index.is_unique


def test_data_Yaws():
    Tc_sum = _crit_Yaws['Tc'].sum()
    assert_allclose(Tc_sum, 5862006.9500000002)

    Pc_sum = _crit_Yaws['Pc'].sum()
    assert_allclose(Pc_sum, 62251189000.0)

    Vc_sum = _crit_Yaws['Vc'].sum()
    assert_allclose(Vc_sum, 4.65511199)

    Zc_sum = _crit_Yaws['Zc'].sum()
    assert_allclose(Zc_sum, 1859.6389176846883, 1e-6)

    omega_sum = _crit_Yaws['omega'].sum()
    assert_allclose(omega_sum, 3170.3041999999996)

    assert _crit_Yaws.shape == (7549, 6)
    assert _crit_Yaws.index.is_unique


def test_relationships():
    Vc = Ihmels(Tc=599.4, Pc=1.19E6)
    Pc = Ihmels(Tc=599.4, Vc=0.00109273333333)
    Tc = Ihmels(Vc=0.00109273333333, Pc=1.19E6)
    assert_allclose([Tc, Pc, Vc], [599.3999999981714, 1190000.0000037064, 0.0010927333333333334])
    with pytest.raises(Exception):
        Ihmels(Tc=559.4)

    Vc = Meissner(Tc=599.4, Pc=1.19E6)
    Pc = Meissner(Tc=599.4, Vc=0.0010695726588235296)
    Tc = Meissner(Vc=0.0010695726588235296, Pc=1.19E6)
    assert_allclose([Tc, Pc, Vc], [599.4000000000001, 1190000.0, 0.0010695726588235296])
    with pytest.raises(Exception):
        Meissner(Tc=559.4)

    Vc = Grigoras(Tc=599.4, Pc=1.19E6)
    Pc = Grigoras(Tc=599.4, Vc=0.00134532)
    Tc = Grigoras(Vc=0.00134532, Pc=1.19E6)
    assert_allclose([Tc, Pc, Vc], [599.4, 1190000.0, 0.00134532])
    with pytest.raises(Exception):
        Grigoras(Tc=559.4)

    Vc1 = critical_surface(Tc=599.4, Pc=1.19E6, Method='IHMELS')
    Vc2 = critical_surface(Tc=599.4, Pc=1.19E6, Method='MEISSNER')
    Vc3 = critical_surface(Tc=599.4, Pc=1.19E6, Method='GRIGORAS')
    assert_allclose([Vc1, Vc2, Vc3],
                    [0.0010927333333333334, 0.0010695726588235296, 0.00134532])

    methods = critical_surface(Tc=599.4, Pc=1.19E6, AvailableMethods=True)
    methods.sort()
    methods_listed = ['IHMELS', 'MEISSNER', 'GRIGORAS', 'NONE']
    methods_listed.sort()
    assert methods == methods_listed

    assert None == critical_surface()
    with pytest.raises(Exception):
        critical_surface(Tc=599.4, Pc=1.19E6, Method='FAIL')

@pytest.mark.slow
def test_Tc_main():
    sources = [_crit_IUPAC, _crit_Matthews, _crit_CRC, _crit_PSRKR4, _crit_PassutDanner, _crit_Yaws]
    CASs = set()
    [CASs.update(set(k.index.values)) for k in sources]

    # Use the default method for each chemical in this file
    Tcs = [Tc(i) for i in CASs]
    Tcs_default_sum = pd.Series(Tcs).sum()
    assert_allclose(Tcs_default_sum, 6054504.896122223)

    assert_allclose(514.0, Tc(CASRN='64-17-5'))

    assert_allclose(647.3, Tc(CASRN='7732-18-5', Method='PSRK'))

    assert_allclose(126.2, Tc(CASRN='7727-37-9', Method='MATTHEWS'))

    assert_allclose(467.3661399548533, Tc(CASRN='64-17-5', Method='SURF'))

    methods = Tc(CASRN='98-01-1', AvailableMethods=True)
    assert methods == ['IUPAC', 'PSRK', 'YAWS', 'NONE']

    # Error handling
    assert None == Tc(CASRN='BADCAS')
    # TODO: Only list critical surface as a method if it can be calculated!
    with pytest.raises(Exception):
        Tc(CASRN='98-01-1', Method='BADMETHOD')

@pytest.mark.slow
def test_Pc_main():
    sources = [_crit_IUPAC, _crit_Matthews, _crit_CRC, _crit_PSRKR4, _crit_PassutDanner, _crit_Yaws]
    CASs = set()
    [CASs.update(set(k.index.values)) for k in sources]

    # Use the default method for each chemical in this file
    Pcs = [Pc(i) for i in CASs]
    Pcs_default_sum = pd.Series(Pcs).sum()
    assert_allclose(Pcs_default_sum, 63159160396.183258)

    assert_allclose(6137000.0, Pc(CASRN='64-17-5'))

    assert_allclose(22048321.0, Pc(CASRN='7732-18-5', Method='PSRK'))

    assert_allclose(3394387.5, Pc(CASRN='7727-37-9', Method='MATTHEWS'))

    assert_allclose(6751845.238095238, Pc(CASRN='64-17-5', Method='SURF'))

    methods = Pc(CASRN='98-01-1', AvailableMethods=True)
    assert methods == ['IUPAC', 'PSRK', 'YAWS', 'NONE']

    # Error handling
    assert None == Pc(CASRN='BADCAS')
    # TODO: Only list critical surface as a method if it can be calculated!
    with pytest.raises(Exception):
        Pc(CASRN='98-01-1', Method='BADMETHOD')

@pytest.mark.slow
def test_Vc_main():
    sources = [_crit_IUPAC, _crit_Matthews, _crit_CRC, _crit_PSRKR4, _crit_Yaws]
    CASs = set()
    [CASs.update(set(k.index.values)) for k in sources]

    # Use the default method for each chemical in this file
    Vcs = [Vc(i) for i in CASs]
    Vcs_default_sum = pd.Series(Vcs).sum()
    assert_allclose(Vcs_default_sum, 4.7958940200000004)

    assert_allclose(0.000168, Vc(CASRN='64-17-5'))

    assert_allclose(5.600e-05, Vc(CASRN='7732-18-5', Method='PSRK'))

    assert_allclose(8.950e-05, Vc(CASRN='7727-37-9', Method='MATTHEWS'))

    assert_allclose(0.00018476306394027916, Vc(CASRN='64-17-5', Method='SURF'))

    methods = Vc(CASRN='98-01-1', AvailableMethods=True)
    assert methods == ['PSRK', 'YAWS', 'NONE']

    # Error handling
    assert None == Vc(CASRN='BADCAS')
    # TODO: Only list critical surface as a method if it can be calculated!
    with pytest.raises(Exception):
        Vc(CASRN='98-01-1', Method='BADMETHOD')

@pytest.mark.slow
def test_Zc_main():
    sources = [_crit_IUPAC, _crit_Matthews, _crit_CRC, _crit_PSRKR4, _crit_Yaws]
    CASs = set()
    [CASs.update(set(k.index.values)) for k in sources]

    # Use the default method for each chemical in this file
    Zcs = [Zc(i) for i in CASs]
    Zcs_default_sum = pd.Series(Zcs).sum()
    assert_allclose(Zcs_default_sum, 1930.7388004412558, 1e-6)

    assert_allclose(0.241, Zc(CASRN='64-17-5'))

    assert_allclose(0.22941610667800444, Zc(CASRN='7732-18-5', Method='PSRK'))

    assert_allclose(0.29, Zc(CASRN='7727-37-9', Method='MATTHEWS'))

    assert_allclose(0.24125051446879994, Zc(CASRN='64-17-5', Method='COMBINED'))

    methods = Zc(CASRN='98-01-1', AvailableMethods=True)
    assert methods == ['PSRK', 'YAWS', 'NONE']

    # Error handling
    assert None == Zc(CASRN='BADCAS')
    with pytest.raises(Exception):
        Zc(CASRN='98-01-1', Method='BADMETHOD')


def test_mixing_Tc():
    # Nitrogen-Argon 50/50 mixture
    Tcm =  Li([0.5, 0.5], [126.2, 150.8], [8.95e-05, 7.49e-05])
    assert_allclose(Tcm, 137.40766423357667)

    # example is from [2]_, for:
    # butane/pentane/hexane 0.6449/0.2359/0.1192 mixture, exp: 450.22 K.
    # Its result is identical to that calculated in the article.
    Tcm = Li([0.6449, 0.2359, 0.1192], [425.12, 469.7, 507.6], [0.000255, 0.000313, 0.000371])
    assert_allclose(Tcm, 449.68261498555444)

    with pytest.raises(Exception):
        Li([0.2359, 0.1192], [425.12, 469.7, 507.6], [0.000255, 0.000313, 0.000371])

    # First example is for an acetone/n-butane 50/50 mixture. No point is
    # available to compare the calculated value with, but it is believed
    # correct.
    Tcm = Chueh_Prausnitz_Tc([0.5, 0.5], [508.1, 425.12], [0.000213, 0.000255], [[0, -14.2619], [-14.2619, 0]])
    assert_allclose(Tcm, 457.01862919555555)

    # 2rd example is from [2]_, for:
    # butane/pentane/hexane 0.6449/0.2359/0.1192 mixture, exp: 450.22 K.
    # Its result is identical to that calculated in the article.
    Tcm = Chueh_Prausnitz_Tc([0.6449, 0.2359, 0.1192], [425.12, 469.7, 507.6], [0.000255, 0.000313, 0.000371], [[0, 1.92681, 6.80358], [1.92681, 0, 1.89312], [ 6.80358, 1.89312, 0]])
    assert_allclose(Tcm, 450.1225764723492)
    # 3rd example is from [2]_, for ethylene, Benzene, ethylbenzene. This is
    # similar to but not identical to the result from the article. The
    # experimental point is 486.9 K.
    Tcm = Chueh_Prausnitz_Tc([0.5, 0.447, .053], [282.34, 562.05, 617.15], [0.0001311, 0.000256, 0.000374], [[0, 37.9570, 0], [37.9570, 0, 4.2459], [0, 4.2459, 0]])
    assert_allclose(Tcm, 475.3154572323848)

    with pytest.raises(Exception):
        Chueh_Prausnitz_Tc([0.447, .053], [282.34, 562.05, 617.15], [0.0001311, 0.000256, 0.000374], [[0, 37.9570, 0], [37.9570, 0, 4.2459], [0, 4.2459, 0]])



    # First example is for an acetone/n-butane 50/50 mixture. No point is
    # available to compare the calculated value with, but it is believed
    # correct. Parameters here are from [2]_.
    Tcm = Grieves_Thodos([0.5, 0.5], [508.1, 425.12], [[0, 0.7137], [1.6496, 0]])
    assert_allclose(Tcm, 456.9398283342622)

    # 2rd example is from [2]_, for:
    # butane/pentane/hexane 0.6449/0.2359/0.1192 mixture, exp: 450.22 K.
    # Its result is identical to that calculated in the article.
    Tcm = Grieves_Thodos([0.6449, 0.2359, 0.1192], [425.12, 469.7, 507.6], [[0, 1.2503, 1.516], [0.799807, 0, 1.23843], [0.659633, 0.807474, 0]])
    assert_allclose(Tcm, 450.1839618758971)

    with pytest.raises(Exception):
        Grieves_Thodos([0.5, 0.5], [508.1, None], [[0, 0.7137], [1.6496, 0]])


    # First example is Acetone/butane 50/50 mixture.
    Tcm = modified_Wilson_Tc([0.5, 0.5], [508.1, 425.12],  [[0, 0.8359], [0, 1.1963]])
    assert_allclose(Tcm, 456.59176287162256)

    # 2rd example is from [2]_, for:
    # butane/pentane/hexane 0.6449/0.2359/0.1192 mixture, exp: 450.22 K.
    # Its result is identical to that calculated in the article.
    Tcm = modified_Wilson_Tc([0.6449, 0.2359, 0.1192], [425.12, 469.7, 507.6], [[0, 1.174450, 1.274390], [0.835914, 0, 1.21038], [0.746878, 0.80677, 0]])
    assert_allclose(Tcm, 450.0305966823031)

    with pytest.raises(Exception):
        Tcm = modified_Wilson_Tc([0.5, 0.5], [508.1, None],  [[0, 0.8359], [0, 1.1963]])


def test_mixing_Chueh_Prausnitz_Vc():
    Vcm = Chueh_Prausnitz_Vc([0.4271, 0.5729], [0.000273, 0.000256], [[0, 5.61847], [5.61847, 0]])
    assert_allclose(Vcm, 0.00026620503424517445)

    with pytest.raises(Exception):
        Chueh_Prausnitz_Vc([0.4271], [0.000273, 0.000256], [[0, 5.61847], [5.61847, 0]])

def test_mixing_modified_Wilson_Vc():
    Vcm = modified_Wilson_Vc([0.4271, 0.5729], [0.000273, 0.000256], [[0, 0.6671250], [1.3939900, 0]])
    assert_allclose(Vcm, 0.00026643350327068809)

    with pytest.raises(Exception):
        modified_Wilson_Vc([0.4271], [0.000273, 0.000256], [[0, 0.6671250], [1.3939900, 0]])


def test_third_property():
    with pytest.raises(Exception):
        third_property('141-62-8')
    assert third_property('1410-62-8', V=True) is None





# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from numpy.testing import assert_allclose
import pytest
import numpy as np
import pandas as pd
from thermo.interface import *
from thermo.interface import VDI_TABULAR
from thermo.identifiers import checkCAS

def test_CSP():
    # p-dichloribenzene at 412.15 K, from DIPPR; value differs due to a slight
    # difference in method.
    sigma1 = Brock_Bird(412.15, 447.3, 685, 3.952E6)
    assert_allclose(sigma1, 0.02208448325192495)

    # Chlorobenzene from Poling, as compared with a % error value at 293 K.
    sigma1 = Brock_Bird(293.15, 404.75, 633.0, 4530000.0)
    assert_allclose(sigma1, 0.032985686413713036)

    # TODO: Find parameters where Brock Bird is negative

    # Chlorobenzene from Poling, as compared with a % error value at 293 K.
    sigma1 = Pitzer(293., 633.0, 4530000.0, 0.249)
    assert_allclose(sigma1, 0.03458453513446387)

    sigma1 = Sastri_Rao(293.15, 404.75, 633.0, 4530000.0)
    sigma2 = Sastri_Rao(293.15, 404.75, 633.0, 4530000.0, chemicaltype='alcohol')
    sigma3 = Sastri_Rao(293.15, 404.75, 633.0, 4530000.0, chemicaltype='acid')
    sigmas = [0.03234567739694441, 0.023255104102733407, 0.02558993464948134]
    assert_allclose([sigma1, sigma2, sigma3], sigmas)

    sigma = Zuo_Stenby(293., 633.0, 4530000.0, 0.249)
    assert_allclose(sigma, 0.03345569011871088)

    # 1-butanol, as compared to value in CRC Handbook of 0.02493.
    sigma = Hakim_Steinberg_Stiel(298.15, 563.0, 4414000.0, 0.59, StielPolar=-0.07872)
    assert_allclose(sigma, 0.021907902575190447)

    sigma_calc = Miqueu(300., 340.1, 0.000199, 0.1687)
    assert_allclose(sigma_calc, 0.003474099603581931)

    
def test_Aleem():
    st = Aleem(T=90, MW=16.04246, Tb=111.6, rhol=458.7, Hvap_Tb=510870., Cpl=2465.)
    assert_allclose(st, 0.01669970221165325)

    # Test the "critical" point (where the correlation predicts sigma=0)
    st_at_Tc = Aleem(T=318.8494929006085, MW=16.04246, Tb=111.6, rhol=458.7, 
                     Hvap_Tb=510870.,  Cpl=2465.)
    assert_allclose(st_at_Tc, 0, atol=1E-12)
    
    
def test_REFPROP():
    sigma = REFPROP(298.15, 647.096, -0.1306, 2.471, 0.2151, 1.233)
    assert_allclose(sigma, 0.07205503890847453)


def test_Somayajulu():
    sigma = Somayajulu(300, 647.126, 232.713514, -140.18645, -4.890098)
    assert_allclose(sigma, 0.07166386387996757)


def test_Jasper():
    sigma = Jasper(298.15, 24, 0.0773)
    assert_allclose(sigma, 0.0220675)


def test_data():
    tot = sum([Mulero_Cachadina_data[i].sum() for i in Mulero_Cachadina_data.columns[1:]])
    assert_allclose(tot, 114350.07371931802)

    assert np.all(Mulero_Cachadina_data.columns == [u'Fluid', u'sigma0', u'n0',
                                                    u'sigma1', u'n1', u'sigma2',
                                                    u'n2', u'Tc', u'Tmin',
                                                    u'Tmax'])
    assert Mulero_Cachadina_data.shape == (115, 10)
    assert Mulero_Cachadina_data.index.is_unique

    tot = sum([Jasper_Lange_data[i].sum() for i in Jasper_Lange_data.columns[1:]])
    assert_allclose(tot, 343153.38953333395)

    assert np.all(Jasper_Lange_data.columns == [u'Name', u'a', u'b', u'Tmin', u'Tmax'])
    assert Jasper_Lange_data.shape == (522, 5)
    assert Jasper_Lange_data.index.is_unique

    tot = sum([Somayajulu_data[i].sum() for i in Somayajulu_data.columns[1:]])
    assert_allclose(tot, 38941.199955999997)

    assert np.all(Somayajulu_data.columns == [u'Chemical', u'Tt', u'Tc', u'A', u'B', u'C'])
    assert Somayajulu_data.shape == (64, 6)
    assert Somayajulu_data.index.is_unique

    tot = sum([Somayajulu_data_2[i].sum() for i in Somayajulu_data_2.columns[1:]])
    assert_allclose(tot, 39471.356771000006)
    assert np.all(Somayajulu_data_2.columns == [u'Chemical', u'Tt', u'Tc', u'A', u'B', u'C'])
    assert Somayajulu_data_2.shape == (64, 6)
    assert Somayajulu_data_2.index.is_unique

def test_VDI_PPDS_11_data():
    '''I believe there are no errors here. 
    '''
    assert all([checkCAS(i) for i in VDI_PPDS_11.index])
    tots_calc = [VDI_PPDS_11[i].abs().sum() for i in [u'A', u'B', u'C', u'D', u'E', u'Tc', u'Tm']]
    tots = [18.495069999999998, 336.69950000000006, 6.5941200000000002, 7.7347200000000003, 6.4262199999999998, 150142.28, 56917.699999999997]
    assert_allclose(tots_calc, tots)
    
    assert VDI_PPDS_11.index.is_unique
    assert VDI_PPDS_11.shape == (272, 8)



@pytest.mark.meta_T_dept
def test_SurfaceTension():
    # Ethanol, test as many methods as possible at once
    EtOH = SurfaceTension(Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125, omega=0.635, StielPolar=-0.01266, CASRN='64-17-5')
    EtOH.T_dependent_property(305.)
    methods = EtOH.sorted_valid_methods
    methods.remove(VDI_TABULAR)
    sigma_calcs = [(EtOH.set_user_methods(i), EtOH.T_dependent_property(305.))[1] for i in methods]
    sigma_exp = [0.021222422444285592, 0.02171156653650729, 0.02171156653650729, 0.021462066798796135, 0.02140008, 0.038055725907414066, 0.03739257387107131, 0.02645171690486362, 0.03905907338532845, 0.03670733205970745]

    assert_allclose(sorted(sigma_calcs), sorted(sigma_exp))
    assert_allclose(EtOH.calculate(305., VDI_TABULAR), 0.021533867879206747, rtol=1E-4)

    # Test that methods return None
    sigma_calcs = [(EtOH.set_user_methods(i, forced=True), EtOH.T_dependent_property(5000))[1] for i in EtOH.sorted_valid_methods]
    assert [None]*11 == sigma_calcs

    EtOH.set_user_methods('VDI_TABULAR', forced=True)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.T_dependent_property(700.)

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')

    # Test Aleem
    
    CH4 = SurfaceTension(Tb=111.65, Cpl=2465., Hvap_Tb=510870., MW=16.04246, Vml=3.497e-05)
    assert_allclose(CH4.T_dependent_property(90), 0.016704545538936296)
    
    assert not CH4.test_method_validity(600, 'Aleem')
    assert CH4.test_method_validity(100, 'Aleem')


def test_Winterfeld_Scriven_Davis():
    # The example is from [2]_; all results agree.
    # The original source has not been reviewed.
    # 16.06 mol% n-pentane, 83.94 mol% dichloromethane at 298.15 K.
    # sigmas are 15.47 and 28.77 respectively, rhos 8.61 kmol/m^3 and 15.53 kmol/m^3

    sigma = Winterfeld_Scriven_Davis([0.1606, 0.8394], [0.01547, 0.02877], [8610., 15530.])
    assert_allclose(sigma, 0.024967388450439817)

    with pytest.raises(Exception):
        Winterfeld_Scriven_Davis([0.1606, 0.8394, 0.118], [0.01547, 0.02877], [8610., 15530.])


def test_Diguilio_Teja():
    # Winterfeld_Scriven_Davis example, with Diguilio Teja. Calculated sigmas at Tbs are
    # 0.01424 and 0.02530.
    # Checked and edited 2017-01-30

    sigma = Diguilio_Teja(T=298.15, xs=[0.1606, 0.8394], sigmas_Tb=[0.01424, 0.02530], Tbs=[309.21, 312.95], Tcs=[469.7, 508.0])
    assert_allclose(sigma, 0.025716823875045505)

    with pytest.raises(Exception):
        Diguilio_Teja(T=298.15, xs=[0.1606, 0.8394, 0.118], sigmas_Tb=[0.01424, 0.02530], Tbs=[309.21, 312.95], Tcs=[469.7, 508.0])

    with pytest.raises(Exception):
         Diguilio_Teja(T=501.85, xs=[0.1606, 0.8394], sigmas_Tb=[0.01424, 0.02530], Tbs=[309.21, 312.95], Tcs=[469.7, 508.0])



def test_SurfaceTensionMixture():
    from thermo.chemical import Mixture
    from thermo.interface import SurfaceTensionMixture, DIGUILIOTEJA, SIMPLE, WINTERFELDSCRIVENDAVIS
    m = Mixture(['pentane', 'dichloromethane'], zs=[.1606, .8394], T=298.15)
    SurfaceTensions = [i.SurfaceTension for i in m.Chemicals]
    VolumeLiquids = [i.VolumeLiquid for i in m.Chemicals]
    
    a = SurfaceTensionMixture(MWs=m.MWs, Tbs=m.Tbs, Tcs=m.Tcs, CASs=m.CASs, SurfaceTensions=SurfaceTensions, VolumeLiquids=VolumeLiquids)

    sigma = a.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(sigma, 0.023886472131054423)
    
    sigma = a.calculate(m.T, m.P, m.zs, m.ws, SIMPLE)
    assert_allclose(sigma, 0.025331490604571537)
    
    sigmas = [a.calculate(m.T, m.P, m.zs, m.ws, i) for i in [DIGUILIOTEJA, SIMPLE, WINTERFELDSCRIVENDAVIS]]
    assert_allclose(sigmas, [0.025257338967448677, 0.025331490604571537, 0.023886472131054423])
    
    with pytest.raises(Exception):
        a.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')
    with pytest.raises(Exception):
        a.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from __future__ import division

__all__ = ['has_CoolProp', 'coolprop_dict', 'CP_fluid', 'coolprop_fluids', 
'CoolProp_T_dependent_property', 'PropsSI', 'PhaseSI', 'CP', 'AbstractState']

try:
    from CoolProp.CoolProp import PropsSI, PhaseSI
    import CoolProp.CoolProp as CP
    from CoolProp import AbstractState
    has_CoolProp = True
except ImportError:  # pragma: no cover
    has_CoolProp = False
    PropsSI, PhaseSI, CP, AbstractState = [None, None, None, None]
#has_CoolProp = False # For testing

# All of these can be inputs to the PropsSI function!
coolprop_dict = ['100-41-4', '10024-97-2', '102687-65-0', '106-42-3',
'106-97-8', '106-98-9', '107-46-0', '107-51-7', '107-52-8', '107-83-5',
'108-38-3', '108-88-3', '109-66-0', '110-54-3', '110-82-7', '111-65-9',
'111-84-2', '112-39-0', '112-40-3', '112-61-8', '112-62-9', '112-63-0',
'1120-21-4', '115-07-1', '115-10-6', '115-11-7', '115-25-3', '124-18-5',
'124-38-9', '1333-74-0', '141-62-8', '141-63-9', '142-82-5', '1717-00-6',
'2551-62-4', '2837-89-0', '287-92-3', '29118-24-9', '29118-25-0', '301-00-8',
'306-83-2', '353-36-6', '354-33-6', '406-58-6', '420-46-2', '421-14-7',
'431-63-0', '431-89-0', '460-73-1', '463-58-1', '463-82-1', '540-97-6',
'556-67-2', '590-18-1', '593-53-3', '616-38-6', '624-64-6', '630-08-0',
'64-17-5', '67-56-1', '67-64-1', '690-39-1', '71-43-2', '74-82-8', '74-84-0',
'74-85-1', '74-87-3', '74-98-6', '74-99-7', '7439-90-9', '7440-01-9',
'7440-37-1', '7440-59-7', '7440-63-3', '7446-09-5', '75-10-5', '75-19-4',
'75-28-5', '75-37-6', '75-43-4', '75-45-6', '75-46-7', '75-68-3', '75-69-4',
'75-71-8', '75-72-9', '75-73-0', '754-12-1', '756-13-8', '76-13-1', '76-14-2',
'76-15-3', '76-16-4', '76-19-7', '7664-41-7', '7727-37-9', '7732-18-5',
'7782-39-0', '7782-41-4', '7782-44-7', '7783-06-4', '7789-20-0', '78-78-4',
'811-97-2', '95-47-6']


class CP_fluid(object):
    # Basic object to store constants for a coolprop fluid, much faster than
    # calling coolprop to retrieve the data when needed
    __slots__ = ['Tmin', 'Tmax', 'Pmax', 'has_melting_line', 'Tc', 'Pc', 'Tt',
                 'omega', 'HEOS']
    
    def __deepcopy__(self, memo):
        # AbstractState("HEOS", CAS) has no deep copy;
        # fortunately, none is needed, so we can just return the existing copy
        return self

    def __init__(self, Tmin, Tmax, Pmax, has_melting_line, Tc, Pc, Tt, omega,
                 HEOS):
        self.Tmin = Tmin
        self.Tmax = Tmax
        self.Pmax = Pmax
        self.has_melting_line = has_melting_line
        self.Tc = Tc
        self.Pc = Pc
        self.Tt = Tt
        self.omega = omega
        self.HEOS = HEOS


# Store the propoerties in a dict of CP_fluid instances
coolprop_fluids = {}
if has_CoolProp:
    for CASRN in coolprop_dict:
        HEOS = AbstractState("HEOS", CASRN)
        coolprop_fluids[CASRN] = CP_fluid(Tmin=HEOS.Tmin(), Tmax=HEOS.Tmax(), Pmax=HEOS.pmax(),
                       has_melting_line=HEOS.has_melting_line(), Tc=HEOS.T_critical(), Pc=HEOS.p_critical(),
                       Tt=HEOS.Ttriple(), omega=HEOS.acentric_factor(), HEOS=HEOS)


def CoolProp_T_dependent_property(T, CASRN, prop, phase):
    r'''Calculates a property of a chemical in either the liquid or gas phase
    as a function of temperature only. This means that the property is
    either at 1 atm or along the saturation curve.

    Parameters
    ----------
        T : float
            Temperature of the fluid [K]
        CASRN : str
            CAS number of the fluid
        prop : str
            CoolProp string shortcut for desired property
        phase : str
            Either 'l' or 'g' for liquid or gas properties respectively

    Returns
    -------
        prop : float
            Desired chemical property, [units]

    Notes
    -----
    For liquids above their boiling point, the liquid property is found on the
    saturation line (at higher pressures). Under their boiling point, the
    property is calculated at 1 atm.

    No liquid calculations are permitted above the critical temperature.

    For gases under the chemical's boiling point, the gas property is found
    on the saturation line (at sub-atmospheric pressures). Above the boiling
    point, the property is calculated at 1 atm.

    An exception is raised if the desired CAS is not supported, or if CoolProp
    is not available.

    The list of strings acceptable as an input for property types is:
    http://www.coolprop.org/coolprop/HighLevelAPI.html#table-of-string-inputs-to-propssi-function

    Examples
    --------
    Water at STP according to IAPWS-95

    >>> CoolProp_T_dependent_property(298.15, '7732-18-5', 'D', 'l')
    997.047636760347

    References
    ----------
    .. [1] Bell, Ian H., Jorrit Wronski, Sylvain Quoilin, and Vincent Lemort.
       "Pure and Pseudo-Pure Fluid Thermophysical Property Evaluation and the
       Open-Source Thermophysical Property Library CoolProp." Industrial &
       Engineering Chemistry Research 53, no. 6 (February 12, 2014):
       2498-2508. doi:10.1021/ie4033999. http://www.coolprop.org/
    '''
    if not has_CoolProp:  # pragma: no cover
        raise Exception('CoolProp library is not installed')
    if CASRN not in coolprop_dict:
        raise Exception('CASRN not in list of supported fluids')
    Tc = coolprop_fluids[CASRN].Tc
    T = float(T) # Do not allow custom objects here
    if phase == 'l':
        if T > Tc:
            raise Exception('For liquid properties, must be under the critical temperature.')
        if PhaseSI('T', T, 'P', 101325, CASRN) in [u'liquid', u'supercritical_liquid']:
            return PropsSI(prop, 'T', T, 'P', 101325, CASRN)
        else:
            return PropsSI(prop, 'T', T, 'Q', 0, CASRN)
    elif phase == 'g':
        if PhaseSI('T', T, 'P', 101325, CASRN) == 'gas':
            return PropsSI(prop, 'T', T, 'P', 101325, CASRN)
        else:
            if T < Tc:
                return PropsSI(prop, 'T', T, 'Q', 1, CASRN)
            else:
                # catch supercritical_gas and friends
                return PropsSI(prop, 'T', T, 'P', 101325, CASRN)
    else:
        raise Exception('Error in CoolProp property function')



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''


from thermo import eos
from thermo import eos_mix
from thermo import elements
from thermo import combustion
from thermo import critical
from thermo import triple
from thermo import virial
from thermo import activity
from thermo import vapor_pressure
from thermo import viscosity
from thermo import reaction
from thermo import safety
from thermo import dipole
from thermo import thermal_conductivity
from thermo import phase_change
from thermo import environment
from thermo import identifiers
from thermo import permittivity
from thermo import volume
from thermo import heat_capacity
from thermo import solubility
from thermo import utils
from thermo import lennard_jones
from thermo import law
from thermo import refractivity
from thermo import electrochem
from thermo import interface

from math import exp, log
import os

import warnings

from scipy.optimize import fsolve
import numpy as np
from numpy import linspace
import matplotlib.pyplot as plt

from thermo.identifiers import CASfromAny
#warnings.simplefilter("always") #error for error
warnings.simplefilter("ignore") #error for error


if __name__ == '__main__':
    import doctest
    doctest.testmod(critical)
    doctest.testmod(triple)
    doctest.testmod(virial)
#    doctest.testmod(adsorption)
#    doctest.testmod(activity)
    doctest.testmod(interface)
    doctest.testmod(vapor_pressure)
    doctest.testmod(viscosity)
    doctest.testmod(reaction)
    doctest.testmod(safety)
    doctest.testmod(dipole)
    doctest.testmod(thermal_conductivity)
    doctest.testmod(phase_change)
    doctest.testmod(environment)
    doctest.testmod(identifiers)
    doctest.testmod(permittivity)
    doctest.testmod(volume)
    doctest.testmod(heat_capacity)
    doctest.testmod(solubility)
    doctest.testmod(utils)
    doctest.testmod(lennard_jones)
#    doctest.testmod(law)
    doctest.testmod(refractivity)
    doctest.testmod(elements)
    doctest.testmod(combustion)
#    doctest.testmod(electrochem)
    doctest.testmod(eos)
    doctest.testmod(eos_mix)




# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from . import acentric
from . import activity
from . import chemical
from . import combustion
from . import critical
from . import coolprop
from . import dipole
from . import dippr
from . import datasheet
from . import electrochem
from . import elements
from . import environment
from . import eos
from . import eos_mix
from . import heat_capacity
from . import identifiers
from . import law
from . import lennard_jones
from . import miscdata
from . import permittivity
from . import phase_change
from . import reaction
from . import refractivity
from . import safety
from . import solubility
from . import interface
from . import thermal_conductivity
from . import triple
from . import unifac
from . import utils
from . import vapor_pressure
from . import virial
from . import viscosity
from . import volume


from .acentric import *
from .activity import *
from .chemical import *
from .combustion import *
from .critical import *
from .coolprop import *
from .dipole import *
from .dippr import *
from .datasheet import *
from .electrochem import *
from .elements import *
from .environment import *
from .eos import *
from .eos_mix import *
from .heat_capacity import *
from .identifiers import *
from .law import *
from .lennard_jones import *
from .miscdata import *
from .permittivity import *
from .phase_change import *
from .reaction import *
from .refractivity import *
from .safety import *
from .solubility import *
from .interface import *
from .thermal_conductivity import *
from .triple import *
from .unifac import *
from .utils import *
from .vapor_pressure import *
from .virial import *
from .viscosity import *
from .volume import *


__all__ = ['activity', 'chemical', 'combustion', 'critical',
 'dipole', 'electrochem', 'elements', 'environment', 'eos', 'eos_mix',
 'heat_capacity',  'identifiers', 'law', 'lennard_jones',
 'miscdata',
 'permittivity', 'phase_change', 'reaction',
 'refractivity', 'safety', 'solubility', 'interface',
 'thermal_conductivity', 'triple', 'utils',
 'vapor_pressure', 'virial', 'viscosity', 'volume', 'acentric', 'coolprop', 
 'datasheet', 'dippr', 'unifac']


__all__.extend(acentric.__all__)
__all__.extend(activity.__all__)
__all__.extend(chemical.__all__)
__all__.extend(combustion.__all__)
__all__.extend(critical.__all__)
__all__.extend(coolprop.__all__)
__all__.extend(dipole.__all__)
__all__.extend(dippr.__all__)
__all__.extend(datasheet.__all__)
__all__.extend(electrochem.__all__)
__all__.extend(elements.__all__)
__all__.extend(environment.__all__)
__all__.extend(eos.__all__)
__all__.extend(eos_mix.__all__)
__all__.extend(heat_capacity.__all__)
__all__.extend(identifiers.__all__)
__all__.extend(law.__all__)
__all__.extend(lennard_jones.__all__)
__all__.extend(miscdata.__all__)
__all__.extend(permittivity.__all__)
__all__.extend(phase_change.__all__)
__all__.extend(reaction.__all__)
__all__.extend(refractivity.__all__)
__all__.extend(safety.__all__)
__all__.extend(solubility.__all__)
__all__.extend(interface.__all__)
__all__.extend(thermal_conductivity.__all__)
__all__.extend(triple.__all__)
__all__.extend(utils.__all__)
__all__.extend(unifac.__all__)
__all__.extend(vapor_pressure.__all__)
__all__.extend(virial.__all__)
__all__.extend(viscosity.__all__)
__all__.extend(volume.__all__)


__version__ = '0.1.34'



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from __future__ import division

__all__ = ['EQ100', 'EQ101', 'EQ102', 'EQ104', 'EQ105', 'EQ106', 'EQ107', 
           'EQ114', 'EQ115', 'EQ116', 'EQ127']

from thermo.utils import log, exp, sinh, cosh


def EQ100(T, A=0, B=0, C=0, D=0, E=0, F=0, G=0):
    r'''DIPPR Equation # 100. Used in calculating the molar heat capacities
    of liquids and solids, liquid thermal conductivity, and solid density.
    All parameters default to zero. As this is a straightforward polynomial,
    no restrictions on parameters apply. Note that high-order polynomials like
    this may need large numbers of decimal places to avoid unnecessary error.

    .. math::
        Y = A + BT + CT^2 + DT^3 + ET^4 + FT^5 + GT^6

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-G : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water liquid heat capacity; DIPPR coefficients normally listed in J/kmol/K.

    >>> EQ100(300, 276370., -2090.1, 8.125, -0.014116, 0.0000093701)
    75355.81

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return A + B*T + C*T**2 + D*T**3 + E*T**4 + F*T**5 + G*T**6


def EQ101(T, A, B, C, D, E):
    r'''DIPPR Equation # 101. Used in calculating vapor pressure, sublimation
    pressure, and liquid viscosity.
    All 5 parameters are required. E is often an integer. As the model is
    exponential, a sufficiently high temperature will cause an OverflowError.
    A negative temperature (or just low, if fit poorly) may cause a math domain
    error.

    .. math::
        Y = \exp\left(A + \frac{B}{T} + C\cdot \ln T + D \cdot T^E\right)

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-E : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water vapor pressure; DIPPR coefficients normally listed in Pa.

    >>> EQ101(300, 73.649, -7258.2, -7.3037, 4.1653E-6, 2)
    3537.44834545549

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return exp(A + B/T + C*log(T) + D*T**E)


def EQ102(T, A, B, C, D):
    r'''DIPPR Equation # 102. Used in calculating vapor viscosity, vapor
    thermal conductivity, and sometimes solid heat capacity. High values of B
    raise an OverflowError.
    All 4 parameters are required. C and D are often 0.

    .. math::
        Y = \frac{A\cdot T^B}{1 + \frac{C}{T} + \frac{D}{T^2}}

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-D : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water vapor viscosity; DIPPR coefficients normally listed in Pa*S.

    >>> EQ102(300, 1.7096E-8, 1.1146, 0, 0)
    9.860384711890639e-06

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return A*T**B/(1. + C/T + D/(T*T))


def EQ104(T, A, B, C, D, E):
    r'''DIPPR Equation #104. Often used in calculating second virial
    coefficients of gases. All 5 parameters are required.
    C, D, and E are normally large values.

    .. math::
        Y = A + \frac{B}{T} + \frac{C}{T^3} + \frac{D}{T^8} + \frac{E}{T^9}

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-E : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water second virial coefficient; DIPPR coefficients normally dimensionless.

    >>> EQ104(300, 0.02222, -26.38, -16750000, -3.894E19, 3.133E21)
    -1.1204179007265151

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return A + B/T + C/T**3. + D/T**8. + E/T**9.


def EQ105(T, A, B, C, D):
    r'''DIPPR Equation #105. Often used in calculating liquid molar density.
    All 4 parameters are required. C is sometimes the fluid's critical
    temperature.

    .. math::
        Y = \frac{A}{B^{1 + (1-\frac{T}{C})^D}}

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-D : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Hexane molar density; DIPPR coefficients normally in kmol/m^3.

    >>> EQ105(300., 0.70824, 0.26411, 507.6, 0.27537)
    7.593170096339236

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return A/B**(1 + (1-T/C)**D)


def EQ106(T, Tc, A, B, C=0, D=0, E=0):
    r'''DIPPR Equation #106. Often used in calculating liquid surface tension,
    and heat of vaporization.
    Only parameters A and B parameters are required; many fits include no
    further parameters. Critical temperature is also required.

    .. math::
        Y = A(1-T_r)^{B + C T_r + D T_r^2 + E T_r^3}

        Tr = \frac{T}{Tc}

    Parameters
    ----------
    T : float
        Temperature, [K]
    Tc : float
        Critical temperature, [K]
    A-D : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water surface tension; DIPPR coefficients normally in Pa*S.

    >>> EQ106(300, 647.096, 0.17766, 2.567, -3.3377, 1.9699)
    0.07231499373541

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    Tr = T/Tc
    return A*(1.-Tr)**(B + C*Tr + D*Tr**2. + E*Tr**3.)


def EQ107(T, A=0, B=0, C=0, D=0, E=0):
    r'''DIPPR Equation #107. Often used in calculating ideal-gas heat capacity.
    All 5 parameters are required.
    Also called the Aly-Lee equation.

    .. math::
        Y = A + B\left[\frac{C/T}{\sinh(C/T)}\right]^2 + D\left[\frac{E/T}{
        \cosh(E/T)}\right]^2

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-E : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water ideal gas molar heat capacity; DIPPR coefficients normally in
    J/kmol/K

    >>> EQ107(300., 33363., 26790., 2610.5, 8896., 1169.)
    33585.90452768923

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    .. [2] Aly, Fouad A., and Lloyd L. Lee. "Self-Consistent Equations for
       Calculating the Ideal Gas Heat Capacity, Enthalpy, and Entropy." Fluid
       Phase Equilibria 6, no. 3 (January 1, 1981): 169-79.
       doi:10.1016/0378-3812(81)85002-9.
    '''
    return A + B*((C/T)/sinh(C/T))**2 + D*((E/T)/cosh(E/T))**2


def EQ114(T, Tc, A, B, C, D):
    r'''DIPPR Equation #114. Rarely used, normally as an alternate liquid
    heat capacity expression. All 4 parameters are required, as well as
    critical temperature.

    .. math::
        Y = \frac{A^2}{\tau} + B - 2AC\tau - AD\tau^2 - \frac{1}{3}C^2\tau^3
        - \frac{1}{2}CD\tau^4 - \frac{1}{5}D^2\tau^5

        \tau = 1 - \frac{T}{Tc}

    Parameters
    ----------
    T : float
        Temperature, [K]
    Tc : float
        Critical temperature, [K]
    A-D : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Hydrogen liquid heat capacity; DIPPR coefficients normally in J/kmol/K.

    >>> EQ114(20, 33.19, 66.653, 6765.9, -123.63, 478.27)
    19423.948911676463

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    t = 1.-T/Tc
    return A**2./t + B - 2.*A*C*t - A*D*t**2. - C**2.*t**3./3. - C*D*t**4./2. - D**2*t**5./5.


def EQ115(T, A, B, C=0, D=0, E=0):
    r'''DIPPR Equation #115. No major uses; has been used as an alternate
    liquid viscosity expression, and as a model for vapor pressure.
    Only parameters A and B are required.

    .. math::
        Y = \exp\left(A + \frac{B}{T} + C\log T + D T^2 + \frac{E}{T^2}\right)

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-E : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    No coefficients found for this expression.

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return exp(A+B/T+C*log(T)+D*T**2 + E/T**2)


def EQ116(T, Tc, A, B, C, D, E):
    r'''DIPPR Equation #116. Used to describe the molar density of water fairly
    precisely; no other uses listed. All 5 parameters are needed, as well as
    the critical temperature.

    .. math::
        Y = A + B\tau^{0.35} + C\tau^{2/3} + D\tau + E\tau^{4/3}

        \tau = 1 - \frac{T}{Tc}

    Parameters
    ----------
    T : float
        Temperature, [K]
    Tc : float
        Critical temperature, [K]
    A-E : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Water liquid molar density; DIPPR coefficients normally in kmol/m^3.

    >>> EQ116(300., 647.096, 17.863, 58.606, -95.396, 213.89, -141.26)
    55.17615446406527

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    tau = 1-T/Tc
    return A + B*tau**0.35 + C*tau**(2/3.) + D*tau + E*tau**(4/3.)


def EQ127(T, A, B, C, D, E, F, G):
    r'''DIPPR Equation #127. Rarely used, and then only in calculating
    ideal-gas heat capacity. All 7 parameters are required.

    .. math::
        Y = A+B\left[\frac{\left(\frac{C}{T}\right)^2\exp\left(\frac{C}{T}
        \right)}{\left(\exp\frac{C}{T}-1 \right)^2}\right]
        +D\left[\frac{\left(\frac{E}{T}\right)^2\exp\left(\frac{E}{T}\right)}
        {\left(\exp\frac{E}{T}-1 \right)^2}\right]
        +F\left[\frac{\left(\frac{G}{T}\right)^2\exp\left(\frac{G}{T}\right)}
        {\left(\exp\frac{G}{T}-1 \right)^2}\right]

    Parameters
    ----------
    T : float
        Temperature, [K]
    A-G : float
        Parameter for the equation; chemical and property specific [-]

    Returns
    -------
    Y : float
        Property [constant-specific]

    Examples
    --------
    Ideal gas heat capacity of methanol; DIPPR coefficients normally in
    J/kmol/K

    >>> EQ127(20., 3.3258E4, 3.6199E4, 1.2057E3, 1.5373E7, 3.2122E3, -1.5318E7, 3.2122E3)
    33258.0

    References
    ----------
    .. [1] Design Institute for Physical Properties, 1996. DIPPR Project 801
       DIPPR/AIChE
    '''
    return A+B*((C/T)**2*exp(C/T)/(exp(C/T) - 1)**2) + \
        D*((E/T)**2*exp(E/T)/(exp(E/T)-1)**2) + \
        F*((G/T)**2*exp(G/T)/(exp(G/T)-1)**2)



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from numpy.testing import assert_allclose
import pytest
import numpy as np
import pandas as pd
from thermo.environment import *


def test_GWP_data():
    dat_calc = [GWP_data[i].sum() for i in [u'Lifetime, years', u'Radiative efficienty, W/m^2/ppb', u'SAR 100yr', u'20yr GWP', u'100yr GWP', u'500yr GWP']]
    dat = [85518.965000000011, 17.063414000000002, 128282.0, 288251, 274671.70000000001, 269051.29999999999]
    assert_allclose(dat_calc, dat)


def test_GWP():
    GWP1_calc = GWP(CASRN='74-82-8')
    GWP2_calc = GWP(CASRN='74-82-8', Method='IPCC (2007) 100yr-SAR')
    assert [GWP1_calc, GWP2_calc] == [25.0, 21.0]

    GWP_available = GWP(CASRN='56-23-5', AvailableMethods=True)
    assert GWP_available == ['IPCC (2007) 100yr', 'IPCC (2007) 100yr-SAR', 'IPCC (2007) 20yr', 'IPCC (2007) 500yr', 'NONE']
    tot = pd.DataFrame( [GWP(i, Method=j) for i in GWP_data.index for j in GWP(i, AvailableMethods=True)]).sum()
    assert_allclose(tot, 960256)

    with pytest.raises(Exception):
        GWP(CASRN='74-82-8', Method='BADMETHOD')


def test_logP_data():
    tot = np.abs(CRClogPDict['logP']).sum()
    assert_allclose(tot, 1216.99)
    assert CRClogPDict.index.is_unique

    tot = np.abs(SyrresDict2['logP']).sum()
    assert_allclose(tot, 25658.06)
    assert SyrresDict2.index.is_unique


def test_logP():
    vals = logP('67-56-1'), logP('124-18-5'), logP('7732-18-5')
    assert_allclose(vals, [-0.74, 6.25, -1.38])

    tot_CRC = np.sum(np.abs(np.array([logP(i, Method='CRC') for i in CRClogPDict.index])))
    assert_allclose(tot_CRC, 1216.99)

    tot_SYRRES = np.sum(np.abs(np.array([logP(i, Method='SYRRES') for i in SyrresDict2.index])))
    assert_allclose(tot_SYRRES, 25658.060000000001)

    with pytest.raises(Exception):
        logP(CASRN='74-82-8', Method='BADMETHOD')

    logP_available = logP('110-54-3', AvailableMethods=True)
    assert logP_available == ['CRC', 'SYRRES', 'NONE']

    assert logP('1124321250-54-3') == None


def test_ODP_data():

    dat_calc = [ODP_data[i].sum() for i in ['ODP2 Max', 'ODP2 Min', 'ODP1 Max', 'ODP1 Min', 'ODP2 Design', 'ODP1 Design', 'Lifetime']]
    dat = [77.641999999999996, 58.521999999999998, 64.140000000000001, 42.734000000000002, 63.10509761272651, 47.809027930358717, 2268.1700000000001]
    assert_allclose(dat_calc, dat)

    assert ODP_data.index.is_unique


def test_ODP():
    V1 = ODP(CASRN='460-86-6')
    V2 = ODP(CASRN='76-14-2', Method='ODP2 Max')
    V3 = ODP(CASRN='76-14-2', Method='ODP1 Max')
    assert_allclose([V1, V2, V3], [7.5, 0.58, 1.0])

    assert ODP(CASRN='148875-98-3', Method='ODP2 string') == '0.2-2.1'

    methods = ['ODP2 Max', 'ODP1 Max', 'ODP2 logarithmic average', 'ODP1 logarithmic average', 'ODP2 Min', 'ODP1 Min', 'ODP2 string', 'ODP1 string', 'NONE']

    assert methods == ODP(CASRN='148875-98-3', AvailableMethods=True)

    with pytest.raises(Exception):
        ODP(CASRN='148875-98-3', Method='BADMETHOD')

    assert ODP(CASRN='14882353275-98-3') == None

    dat_calc = [pd.to_numeric(pd.Series([ODP(i, Method=j) for i in ODP_data.index]), errors='coerce').sum() for j in ODP_methods]

    dat = [77.641999999999996, 64.140000000000001, 63.10509761272651, 47.809027930358717, 58.521999999999998, 42.734000000000002, 54.342000000000006, 38.280000000000001]
    assert_allclose(dat_calc, dat)


# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from numpy.testing import assert_allclose
import pytest
from thermo.thermal_conductivity import *
from thermo.chemical import Mixture
from thermo.identifiers import checkCAS
from thermo.thermal_conductivity import MAGOMEDOV, DIPPR_9H, FILIPPOV, SIMPLE, ThermalConductivityLiquidMixture
from thermo.thermal_conductivity import (GHARAGHEIZI_G, CHUNG, ELI_HANLEY, VDI_PPDS,
                                        ELI_HANLEY_DENSE, CHUNG_DENSE, 
                                        EUCKEN_MOD, EUCKEN, BAHADORI_G, 
                                        STIEL_THODOS_DENSE, DIPPR_9B, COOLPROP,
                                        DIPPR_PERRY_8E, VDI_TABULAR, GHARAGHEIZI_L,
                                       SATO_RIEDEL, NICOLA, NICOLA_ORIGINAL,
                                       SHEFFY_JOHNSON, BAHADORI_L,
                                       LAKSHMI_PRASAD, DIPPR_9G, MISSENARD)

def test_Perrys2_314_data():
    # In perry's, only 102 is used. No chemicals are missing.
    # Tmaxs all match to 5E-4. Tmins match to 1E-3.
    assert all([checkCAS(i) for i in Perrys2_314.index])
    tots_calc = [Perrys2_314[i].abs().sum() for i in [u'C1', u'C2', u'C3', u'C4', u'Tmin', u'Tmax']]
    tots = [48935634.823768869, 297.41545078799999, 421906466448.71423, 232863514627157.62, 125020.26000000001, 347743.42000000004]
    assert_allclose(tots_calc, tots)
    
    assert Perrys2_314.index.is_unique
    assert Perrys2_314.shape == (345, 7)


def test_Perrys2_315_data():
    # From perry's - Deuterium  ,  Nitrogen trifluoride  ,  Nitrous oxide   Silicon tetrafluoride  ,  Terephthalic acid  all have no data
    # All perry's use #100.
    # Tmins all match at 5E-4.
    # Tmaxs all match at 2E-3.
    assert all([checkCAS(i) for i in Perrys2_315.index])
    tots_calc = [Perrys2_315[i].abs().sum() for i in [u'C1', u'C2', u'C3', u'C4', u'C5', u'Tmin', u'Tmax']]
    tots = [82.001667499999996, 0.19894598900000002, 0.0065330144999999999, 0.00046928630199999995, 1.0268010799999999e-07, 70996.369999999995, 138833.41]
    assert_allclose(tots_calc, tots)
    
    assert Perrys2_315.index.is_unique
    assert Perrys2_315.shape == (340, 8)


def test_VDI_PPDS_10_data():
    '''Average deviation of 2.4% from tabulated values. Many chemicals have
    much higher deviations. 10% or more deviations:
    ['75-34-3', '107-06-2', '106-93-4', '420-46-2', '71-55-6', '79-34-5', 
    '67-72-1', '76-12-0', '76-13-1', '76-14-2', '540-54-5', '75-01-4', 
    '75-35-4', '79-01-6', '127-18-4', '462-06-6', '108-90-7', '108-86-1', 
    '108-41-8', '100-44-7', '108-93-0', '100-61-8', '121-69-7', '91-66-7']
    
    These have been checked - it appears the tabulated data is just incorrect.
    '''

    assert all([checkCAS(i) for i in VDI_PPDS_10.index])
    tots_calc = [VDI_PPDS_10[i].abs().sum() for i in [u'A', u'B', u'C', u'D', u'E']]
    tots = [2.2974640014599998, 0.015556001460000001, 1.9897655000000001e-05, 6.7747269999999993e-09, 2.3260109999999999e-12]
    assert_allclose(tots_calc, tots)
    
    assert VDI_PPDS_10.index.is_unique
    assert VDI_PPDS_10.shape == (275, 6)


def test_VDI_PPDS_9_data():
    '''Average deviation of 0.71% from tabulated values. The following have 
    larger deviations
        
    ['124-18-5', '629-59-4', '629-78-7', '526-73-8', '95-63-6']
    
    These have been checked - it appears the tabulated data is just incorrect.
    '''

    assert all([checkCAS(i) for i in VDI_PPDS_9.index])
    tots_calc = [VDI_PPDS_9[i].abs().sum() for i in [u'A', u'B', u'C', u'D', u'E']]
    tots = [63.458699999999993, 0.14461469999999998, 0.00042270770000000005, 1.7062660000000002e-06, 3.2715370000000003e-09]
    assert_allclose(tots_calc, tots)
    
    assert VDI_PPDS_9.index.is_unique
    assert VDI_PPDS_9.shape == (271, 6)


def test_CSP_liq():
    kl = Sheffy_Johnson(300, 47, 280)
    assert_allclose(kl, 0.17740150413112196)

    kl = Sato_Riedel(300, 47, 390, 520)
    assert_allclose(kl, 0.2103769246133769)

    kl = Lakshmi_Prasad(273.15, 100)
    assert_allclose(kl, 0.013664450000000009)

    kl = Gharagheizi_liquid(300, 40, 350, 1E6, 0.27)
    assert_allclose(kl, 0.2171113029534838)

    kl = Nicola_original(300, 142.3, 611.7, 0.49, 201853)
    assert_allclose(kl, 0.2305018632230984)

    kl = Nicola(300, 142.3, 611.7, 2110000.0, 0.49)
    assert_allclose(kl, 0.10863821554584034)
    # Not at all sure about this one

    kl = Bahadori_liquid(273.15, 170)
    assert_allclose(kl, 0.14274278108272603)


def test_CSP_liq_dense():
    # From [2]_, for butyl acetate.
    kl_dense = DIPPR9G(515.05, 3.92E7, 579.15, 3.212E6, 7.085E-2)
    assert_allclose(kl_dense, 0.0864419738671184)

    kld1 = Missenard(304., 6330E5, 591.8, 41E5, 0.129)
    # # butyl acetate
    kld2 = Missenard(515.05, 3.92E7, 579.15, 3.212E6, 7.085E-2)
    assert_allclose([kld1, kld2], [0.21983757770696569, 0.086362465280714396])


def test_CSP_gas():
    # 2-methylbutane at low pressure, 373.15 K. Mathes calculation in [1]_.
    kg =  Eucken(72.151, 135.9, 8.77E-6)
    assert_allclose(kg, 0.018792644287722975)

    # 2-methylbutane at low pressure, 373.15 K. Mathes calculation in [1]_.
    kg = Eucken_modified(72.151, 135.9, 8.77E-6)
    assert_allclose(kg, 0.023593536999201956)

    # CO, brute force tests on three  options for chemtype
    kg1 = DIPPR9B(200., 28.01, 20.826, 1.277E-5, 132.92, chemtype='linear')
    assert kg1 == DIPPR9B(200., 28.01, 20.826, 1.277E-5, 132.92) # No argument
    kg2 = DIPPR9B(200., 28.01, 20.826, 1.277E-5, 132.92, chemtype='monoatomic')
    kg3 = DIPPR9B(200., 28.01, 20.826, 1.277E-5, 132.92, chemtype='nonlinear')
    assert_allclose([kg1, kg2, kg3], [0.01813208676438415, 0.023736881470903245, 0.018625352738307743])

    with pytest.raises(Exception):
        DIPPR9B(200., 28.01, 20.826, 1.277E-5, 132.92, chemtype='FAIL')


    kg = Chung(T=373.15, MW=72.151, Tc=460.4, omega=0.227, Cvm=135.9, mu=8.77E-6)
    assert_allclose(kg, 0.023015653729496946)

    kg = eli_hanley(T=373.15, MW=72.151, Tc=460.4, Vc=3.06E-4, Zc=0.267, omega=0.227, Cvm=135.9)
    assert_allclose(kg, 0.022479517891353377)

    kg = eli_hanley(T=1000, MW=72.151, Tc=460.4, Vc=3.06E-4, Zc=0.267, omega=0.227, Cvm=135.9)
    assert_allclose(kg, 0.06369581356766069)

    kg = Bahadori_gas(40+273.15, 20)
    assert_allclose(kg, 0.031968165337873326)

    kg = Gharagheizi_gas(580., 16.04246, 111.66, 4599000.0, 0.0115478000)
    assert_allclose(kg, 0.09594861261873211)


def test_CSP_gas_dense():
    kgs = [stiel_thodos_dense(T=378.15, MW=44.013, Tc=309.6, Pc=72.4E5, Vc=97.4E-6, Zc=0.274, Vm=i, kg=2.34E-2) for i in [144E-6, 24E-6, 240E-6]]
    kgs_exp = [0.041245574404863684, 0.9158718777539487, 0.03258313269922979]
    assert_allclose(kgs, kgs_exp)


    kgs = [eli_hanley_dense(T=T, MW=42.081, Tc=364.9, Vc=1.81E-4, Zc=0.274, omega=0.144, Cvm=82.70, Vm=1.721E-4) for T in [473., 900]]
    kgs_exp = [0.06038475936515042, 0.08987438807653142]
    assert_allclose(kgs, kgs_exp)

    kg = eli_hanley_dense(700, MW=42.081, Tc=364.9, Vc=1.81E-4, Zc=0.274, omega=0.144, Cvm=82.70, Vm=3.721E-4)
    assert_allclose(kg, 0.06953791121177173)

    kg = chung_dense(T=473., MW=42.081, Tc=364.9, Vc=184.6E-6, omega=0.142, Cvm=82.67, Vm=172.1E-6, mu=134E-7, dipole=0.4)
    assert_allclose(kg, 0.06160570379787278)


@pytest.mark.meta_T_dept
def test_ThermalConductivityLiquid():
    EtOH = ThermalConductivityLiquid(CASRN='64-17-5', MW=46.06844, Tm=159.05, Tb=351.39, Tc=514.0, Pc=6137000.0, omega=0.635, Hfus=4931.0)

    EtOH.T_dependent_property(305.)
    methods = list(EtOH.sorted_valid_methods)
    methods.remove(VDI_TABULAR)
    kl_calcs = [(EtOH.set_user_methods(i), EtOH.T_dependent_property(305.))[1] for i in methods]
    kl_exp = [0.162183005823234, 0.16627999999999998, 0.166302, 0.20068212675966418, 0.18526367184633258, 0.18846433785041306, 0.16837295487233528, 0.16883011582627103, 0.09330268101157643, 0.028604363267557775]
    assert_allclose(sorted(kl_calcs), sorted(kl_exp))
    
    assert_allclose(EtOH.calculate(305., VDI_TABULAR), 0.17417420086033197, rtol=1E-4)
    

    # Test that methods return None
    kl_calcs = [(EtOH.set_user_methods(i, forced=True), EtOH.T_dependent_property(5000))[1] for i in EtOH.sorted_valid_methods]
    assert [None]*11 == kl_calcs

    EtOH.set_user_methods(VDI_TABULAR, forced=True)
    assert_allclose(EtOH.T_dependent_property(600.), 0.040117737789202995)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.T_dependent_property(600.)

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')


    # Ethanol compressed
    assert [False, True] == [EtOH.test_method_validity_P(300, P, COOLPROP) for P in (1E3, 1E5)]
    assert [True, True] == [EtOH.test_method_validity_P(300, P, DIPPR_9G) for P in (1E3, 1E5)]
    assert [True, True, False] == [EtOH.test_method_validity_P(300, P, MISSENARD) for P in (1E3, 1E5, 1E10)]

    EtOH = ThermalConductivityLiquid(CASRN='64-17-5', MW=46.06844, Tm=159.05, Tb=351.39, Tc=514.0, Pc=6137000.0, omega=0.635, Hfus=4931.0)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, COOLPROP), 0.1639626989794703)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, DIPPR_9G), 0.1606146938795702)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, MISSENARD), 0.1641582259181843)


    # Ethanol data, calculated from CoolProp
    Ts = [275, 300, 350]
    Ps = [1E5, 5E5, 1E6]
    TP_data = [[0.16848555706973622, 0.16313525757474362, 0.15458068887966378], [0.16868861153075654, 0.163343255114212, 0.1548036152853355], [0.16894182645698885, 0.1636025336196736, 0.15508116339039268]]
    EtOH.set_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_allclose(TP_data, recalc_pts)

    EtOH.forced_P = True
    assert_allclose(EtOH.TP_dependent_property(274, 9E4), 0.16848555706973622)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.TP_dependent_property(300, 9E4)

    with pytest.raises(Exception):
        EtOH.test_method_validity_P(300, 1E5, 'BADMETHOD')

    assert False == EtOH.test_method_validity_P(-10, 1E5, DIPPR_9G)


@pytest.mark.meta_T_dept
def test_ThermalConductivityGas():
    EtOH = ThermalConductivityGas(MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.2412, omega=0.635, dipole=1.44, Vmg=0.02357, Cvgm=56.98, mug=7.903e-6, CASRN='64-17-5')
    EtOH.T_dependent_property(298.15)
    kg_calcs = [(EtOH.set_user_methods(i), EtOH.T_dependent_property(298.15))[1] for i in EtOH.sorted_valid_methods]
    kg_exp = [0.015227631457903644, 0.015025094773729045, 0.01520257225203181, 0.01494275, 0.016338750949017277, 0.014353317470206847, 0.011676848981094841, 0.01137910777526855, 0.015427444948536088, 0.012984129385510995, 0.017556325226536728]
    assert_allclose(kg_calcs, kg_exp)

    # Test that those mthods which can, do, return NoneEtOH.forced_P
    kg_calcs = [(EtOH.set_user_methods(i, forced=True), EtOH.T_dependent_property(5E20))[1] for i in [COOLPROP, DIPPR_PERRY_8E, VDI_TABULAR, GHARAGHEIZI_G, ELI_HANLEY, BAHADORI_G, VDI_PPDS]]
    assert [None]*7 == kg_calcs

    # Test tabular limits/extrapolation
    EtOH.set_user_methods(VDI_TABULAR, forced=True)
    assert_allclose(EtOH.T_dependent_property(600.), 0.05755089974293061)

    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.T_dependent_property(600.)

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')


    # Ethanol compressed

    assert [True, False] == [EtOH.test_method_validity_P(300, P, COOLPROP) for P in (1E3, 1E5)]
    assert [True, False] == [EtOH.test_method_validity_P(300, P, ELI_HANLEY_DENSE) for P in (1E5, -1E5)]
    assert [True, False] == [EtOH.test_method_validity_P(300, P, CHUNG_DENSE) for P in (1E5, -1E5)]
    assert [True, False] == [EtOH.test_method_validity_P(300, P, STIEL_THODOS_DENSE) for P in (1E5, -1E5)]


    EtOH = ThermalConductivityGas(MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.2412, omega=0.635, dipole=1.44, Vmg=0.02357, Cvgm=56.98, mug=7.903e-6, CASRN='64-17-5')
    assert_allclose(EtOH.calculate_P(298.15, 1E2, COOLPROP), 0.015207849649231962)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, ELI_HANLEY_DENSE), 0.011210125242396791)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, CHUNG_DENSE), 0.011770372068085207)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, STIEL_THODOS_DENSE), 0.015447836685420897)


      # Ethanol data, calculated from CoolProp
    Ts = [400, 500, 600]
    Ps = [1E4, 1E5, 2E5]
    TP_data = [[0.025825794817543015, 0.037905383602635095, 0.05080124980338535], [0.02601702567554805, 0.03806794452306919, 0.050946301396380594], [0.026243171168075605, 0.03825284803978187, 0.05110925652065333]]
    EtOH.set_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_allclose(TP_data, recalc_pts)

    EtOH.forced_P = True
    EtOH.tabular_extrapolation_permitted = True
    assert_allclose(EtOH.TP_dependent_property(399, 9E3), 0.025825794817543015)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.TP_dependent_property(399, 9E3)

    with pytest.raises(Exception):
        EtOH.test_method_validity_P(300, 1E5, 'BADMETHOD')

    assert False == EtOH.test_method_validity_P(100, 1E5, COOLPROP)


def test_DIPPR9H():
    k = DIPPR9H([0.258, 0.742], [0.1692, 0.1528])
    assert_allclose(k, 0.15657104706719646)

    with pytest.raises(Exception):
        DIPPR9H([0.258, 0.742], [0.1692])

def test_Filippov():
    kl = Filippov([0.258, 0.742], [0.1692, 0.1528])
    assert_allclose(kl, 0.15929167628799998)

    with pytest.raises(Exception):
        Filippov([0.258], [0.1692, 0.1528])


def test_Lindsay_Bromley():
    kg = Lindsay_Bromley(323.15, [0.23, 0.77], [1.939E-2, 1.231E-2], [1.002E-5, 1.015E-5], [248.31, 248.93], [46.07, 50.49])
    assert_allclose(kg, 0.01390264417969313)

    with pytest.raises(Exception):
        Lindsay_Bromley(323.15, [0.23], [1.939E-2, 1.231E-2], [1.002E-5, 1.015E-5], [248.31, 248.93], [46.07, 50.49])



def test_ThermalConductivityGasMixture():
    from thermo.chemical import Mixture
    from thermo.thermal_conductivity import ThermalConductivityGasMixture, LINDSAY_BROMLEY, SIMPLE
    
    m2 = Mixture(['nitrogen', 'argon', 'oxygen'], ws=[0.7557, 0.0127, 0.2316])
    ThermalConductivityGases = [i.ThermalConductivityGas for i in m2.Chemicals]
    ViscosityGases = [i.ViscosityGas for i in m2.Chemicals]

    kg_mix = ThermalConductivityGasMixture(MWs=m2.MWs, Tbs=m2.Tbs, CASs=m2.CASs, 
                                      ThermalConductivityGases=ThermalConductivityGases, 
                                      ViscosityGases=ViscosityGases)
    
    k = kg_mix.mixture_property(m2.T, m2.P, m2.zs, m2.ws)
    assert_allclose(k, 0.025864474514829254) # test LINDSAY_BROMLEY and mixture property
    # Do it twice to test the stored method
    k = kg_mix.mixture_property(m2.T, m2.P, m2.zs, m2.ws)
    assert_allclose(k, 0.025864474514829254) # test LINDSAY_BROMLEY and mixture property

    k =  kg_mix.calculate(m2.T, m2.P, m2.zs, m2.ws, SIMPLE) # Test calculate, and simple
    assert_allclose(k, 0.02586655464213776)
    
    dT1 = kg_mix.calculate_derivative_T(m2.T, m2.P, m2.zs, m2.ws, LINDSAY_BROMLEY)
    dT2 = kg_mix.property_derivative_T(m2.T, m2.P, m2.zs, m2.ws)
    assert_allclose([dT1, dT2], [7.3391064059347144e-05]*2)
    
    dP1 = kg_mix.calculate_derivative_P(m2.P, m2.T, m2.zs, m2.ws, LINDSAY_BROMLEY)
    dP2 = kg_mix.property_derivative_P(m2.T, m2.P, m2.zs, m2.ws)
    
    assert_allclose([dP1, dP2], [3.5325319058809868e-10]*2, rtol=1E-4)
    
    # Test other methods
    
    assert kg_mix.user_methods == []
    assert kg_mix.all_methods == {LINDSAY_BROMLEY, SIMPLE}
    assert kg_mix.ranked_methods == [LINDSAY_BROMLEY, SIMPLE]
    
    # set a method
    kg_mix.set_user_method([SIMPLE])
    assert None == kg_mix.method
    k = kg_mix.mixture_property(m2.T, m2.P, m2.zs, m2.ws)
    assert_allclose(k, 0.02586655464213776)
    
    # Unhappy paths
    with pytest.raises(Exception):
        kg_mix.calculate(m2.T, m2.P, m2.zs, m2.ws, 'BADMETHOD')
        
    with pytest.raises(Exception):
        kg_mix.test_method_validity(m2.T, m2.P, m2.zs, m2.ws, 'BADMETHOD')



def test_ThermalConductivityLiquidMixture():
    from thermo.thermal_conductivity import MAGOMEDOV, DIPPR_9H, FILIPPOV, SIMPLE, ThermalConductivityLiquidMixture

    m = Mixture(['ethanol', 'pentanol'], ws=[0.258, 0.742], T=298.15)
    ThermalConductivityLiquids = [i.ThermalConductivityLiquid for i in m.Chemicals]
    
    kl_mix = ThermalConductivityLiquidMixture(CASs=m.CASs, ThermalConductivityLiquids=ThermalConductivityLiquids)
    k = kl_mix.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(k, 0.15300152782218343)
                           
    k = kl_mix.calculate(m.T, m.P, m.zs, m.ws, FILIPPOV)
    assert_allclose(k, 0.15522139770330717)
    
    k = kl_mix.calculate(m.T, m.P, m.zs, m.ws, SIMPLE)
    assert_allclose(k, 0.1552717795028546)
    
    # Test electrolytes
    m = Mixture(['water', 'sulfuric acid'], ws=[.5, .5], T=298.15)
    ThermalConductivityLiquids = [i.ThermalConductivityLiquid for i in m.Chemicals]
    kl_mix = ThermalConductivityLiquidMixture(CASs=m.CASs, ThermalConductivityLiquids=ThermalConductivityLiquids)
    k = kl_mix.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(k, 0.4677453168207703)
    assert kl_mix.sorted_valid_methods == [MAGOMEDOV]
                     

    # Unhappy paths
    with pytest.raises(Exception):
        kl_mix.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')
        
    with pytest.raises(Exception):
        kl_mix.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from numpy.testing import assert_allclose
import pytest
import numpy as np
import pandas as pd
from thermo.viscosity import *
from thermo.identifiers import checkCAS
from thermo.viscosity import COOLPROP, LUCAS
from thermo.chemical import Mixture
from thermo.viscosity import LALIBERTE_MU, MIXING_LOG_MOLAR, MIXING_LOG_MASS, BROKAW, HERNING_ZIPPERER, WILKE, SIMPLE

### Check data integrity

def test_Dutt_Prasad_data():
    assert all([checkCAS(i) for i in Dutt_Prasad.index])
    tots_calc = [Dutt_Prasad[i].abs().sum() for i in ['A', 'B', 'C', 'Tmin', 'Tmax']]
    tots = [195.89260000000002, 65395.299999999996, 9849.1899999999987, 25952, 35016]
    assert_allclose(tots_calc, tots)

    assert Dutt_Prasad.index.is_unique
    assert Dutt_Prasad.shape == (100, 6)


def test_VN2E_data():
    assert all([checkCAS(i) for i in VN2E_data.index])
    tots_calc = [VN2E_data[i].abs().sum() for i in ['C', 'D', 'Tmin', 'Tmax']]
    tots = [567743298666.74878, 48.8643, 3690, 4860]
    assert_allclose(tots_calc, tots)

    assert VN2E_data.index.is_unique
    assert VN2E_data.shape == (14, 6)


def test_VN3_data():
    assert all([checkCAS(i) for i in VN3_data.index])
    tots_calc = [VN3_data[i].abs().sum() for i in ['A', 'B', 'C', 'Tmin', 'Tmax']]
    tots = [645.18849999999998, 169572.65159999998, 50050.151870000002, 126495, 175660]
    assert_allclose(tots_calc, tots)

    assert VN3_data.index.is_unique
    assert VN3_data.shape == (432, 7)

def test_VN2_data():
    assert all([checkCAS(i) for i in VN3_data.index])
    tots_calc = [VN2_data[i].abs().sum() for i in ['A', 'B', 'Tmin', 'Tmax']]
    tots = [674.10069999999996, 83331.98599999999, 39580, 47897]
    assert_allclose(tots_calc, tots)

    assert VN2_data.index.is_unique
    assert VN2_data.shape == (135, 6)


def test_Perrys2_313_data():
    # All values calculated at Tmin and Tmax check out to at least 5E-3 precision
    # The rounding has some effect, but it is not worrying.
    assert all([checkCAS(i) for i in Perrys2_313.index])
    tots_calc = [Perrys2_313[i].abs().sum() for i in [u'C1', u'C2', u'C3', u'C4', u'C5', u'Tmin', u'Tmax']]
    tots = [9166.6971369999992, 615425.94497999991, 1125.5317557875198, 9.054869390623603e+34, 402.21244000000002, 72467.140000000014, 136954.85999999999]
    assert_allclose(tots_calc, tots)
    
    assert Perrys2_313.index.is_unique
    assert Perrys2_313.shape == (337, 8)

    
def test_Perrys2_312_data():
    # Argon, Difluoromethane, 1-Hexyne, Methylsilane, Nitrogen trifluoride, 
    # Vinyl chloride all do not match on Tmax at 1E-3 - their models predict 
    # ~1E-5 Pa*S, but listed values are ~1E-10 to 1E-12. Unsure of the cause.
    # All coumpounds match at 1E-3 for Tmin.
    
    assert all([checkCAS(i) for i in Perrys2_312.index])
    tots_calc = [Perrys2_312[i].abs().sum() for i in [u'C1', u'C2', u'C3', u'C4', u'Tmin', u'Tmax']]
    tots = [0.00019683902626010103, 250.10520100000002, 65862.829200000007, 191286, 74802.639999999999, 355064.37]
    assert_allclose(tots_calc, tots)
    
    assert Perrys2_312.index.is_unique
    assert Perrys2_312.shape == (345, 7)


def test_VDI_PPDS_7_data():
    assert all([checkCAS(i) for i in VDI_PPDS_7.index])
    tots_calc = [VDI_PPDS_7[i].abs().sum() for i in [u'A', u'B', u'C', u'D', u'E']]
    tots = [507.14607000000001, 1680.7624099999998, 165461.14259999999, 46770.887000000002, 0.057384780000000003]
    assert_allclose(tots_calc, tots)
    
    assert VDI_PPDS_7.index.is_unique
    assert VDI_PPDS_7.shape == (271, 7)

def test_VDI_PPDS_8_data():
    # Coefficients for water are incorrect - obtained an average deviation of 150%!
    assert all([checkCAS(i) for i in VDI_PPDS_8.index])
    tots_calc = [VDI_PPDS_8[i].abs().sum() for i in [u'A', u'B', u'C', u'D', u'E']]
    tots = [0.00032879559999999999, 9.5561339999999995e-06, 2.8377710000000001e-09, 2.8713399999999998e-12, 2.8409200000000004e-15]
    assert_allclose(tots_calc, tots)
    
    assert VDI_PPDS_8.index.is_unique
    assert VDI_PPDS_8.shape == (274, 6)

def test_ViswanathNatarajan():
    mu = ViswanathNatarajan2(348.15, -5.9719, 1007.0)
    assert_allclose(mu, 0.00045983686956829517)

    mu = ViswanathNatarajan2Exponential(298.15, 4900800,  -3.8075)
    assert_allclose(mu, 0.0018571903840928496)

    mu = ViswanathNatarajan3(298.15, -2.7173, -1071.18, -129.51)
    assert_allclose(mu, 0.0006129806445142112)


def test_Letsou_Stiel():
    # Checked 2017-03-05
    mu = Letsou_Stiel(400., 46.07, 516.25, 6.383E6, 0.6371)
    assert_allclose(mu, 0.0002036150875308151)

    # Butane, custom case; vs 0.000166 exp
    mu = Letsou_Stiel(298.15, 58.1222, 425.12, 3796000.0, 0.193)
    assert_allclose(mu, 0.00015385075092073157)


def test_Przedziecki_Sridhar():
    # Reid (1983) toluene, 383K
    mu = Przedziecki_Sridhar(383., 178., 591.8, 41E5, 316E-6, 95E-6, .263, 92.14)
    assert_allclose(mu, 0.0002198147995603383)


def test_Lucas():
    # methylcyclohexane
    mu = Lucas(300., 500E5, 572.2, 34.7E5, 0.236, 0, 0.00068)
    assert_allclose(mu, 0.0010683738499316518)

    # Psat > P:
    mu = Lucas(300., 500E5, 572.2, 34.7E5, 0.236, 501E5, 0.00068)
    assert_allclose(mu, 0.00068)


def test_Stiel_Thodos():
    # CCl4
    mu = Stiel_Thodos(300., 556.35, 4.5596E6, 153.8)
    assert_allclose(mu, 1.0408926223608723e-05)

    # Tr > 1.5
    mu = Stiel_Thodos(900., 556.35, 4.5596E6, 153.8)
    assert_allclose(mu, 2.899111242556782e-05)


def test_Yoon_Thodos():
    # CCl4
    mu = Yoon_Thodos(300., 556.35, 4.5596E6, 153.8)
    assert_allclose(mu, 1.0194885727776819e-05)


def test_Gharagheizi_gas_viscosity():
    mu = Gharagheizi_gas_viscosity(120., 190.564, 45.99E5, 16.04246)
    assert_allclose(mu, 5.215761625399613e-06)


def test_Lucas_gas():
    mu = lucas_gas(T=550., Tc=512.6, Pc=80.9E5, Zc=0.224, MW=32.042, dipole=1.7)
    assert_allclose(mu, 1.7822676912698928e-05)

    mu = lucas_gas(T=550., Tc=512.6, Pc=80.9E5, Zc=0.224, MW=32.042, dipole=None)
    assert_allclose(mu, 1.371116974367763e-05)

    mu = lucas_gas(T=550., Tc=512.6, Pc=80.9E5, Zc=0.224, MW=32.042, dipole=8)
    assert_allclose(mu, 1.7811559961984407e-05)

    # Helium, testing Q
    mu = lucas_gas(T=6, Tc=5.1889, Pc=226968.0, Zc=0.3014, MW=4.002602, CASRN='7440-59-7')
    assert_allclose(mu, 1.3042945737346396e-06)

    mu = lucas_gas(T=150, Tc=5.1889, Pc=226968.0, Zc=0.3014, MW=4.002602, CASRN='7440-59-7')
    assert_allclose(mu, 1.2558477184738118e-05)


def test_Wilke():
    mu = Wilke([0.05, 0.95], [1.34E-5, 9.5029E-6], [64.06, 46.07])
    assert_allclose(mu, 9.701614885866193e-06)

    with pytest.raises(Exception):
        Wilke([0.05], [1.34E-5, 9.5029E-6], [64.06, 46.07])


def test_Herning_Zipperer():
    mu = Herning_Zipperer([0.05, 0.95], [1.34E-5, 9.5029E-6], [64.06, 46.07])
    assert_allclose(mu, 9.730630997268096e-06)

    with pytest.raises(Exception):
        Herning_Zipperer([0.05], [1.34E-5, 9.5029E-6], [64.06, 46.07])


def test_Brockaw():
    mu = Brokaw(308.2, [0.05, 0.95], [1.34E-5, 9.5029E-6], [64.06, 46.07], [0.42, 0.19], [347, 432])
    assert_allclose(mu, 9.699085099801568e-06)

    with pytest.raises(Exception):
        Brokaw(308.2, [0.95], [1.34E-5, 9.5029E-6], [64.06, 46.07], [0.42, 0.19], [347, 432])

    # Test < 0.1 MD
    mu = Brokaw(308.2, [0.05, 0.95], [1.34E-5, 9.5029E-6], [64.06, 46.07], [0.42, 0.05], [347, 432])
    assert_allclose(mu, 9.70504431025103e-06)


def test_round_whole_even():
    from thermo.viscosity import _round_whole_even
    assert _round_whole_even(116.4) == 116
    assert _round_whole_even(116.6) == 117
    assert _round_whole_even(116.5) == 116
    assert _round_whole_even(115.5) == 116


def test_viscosity_index():
    # 4/5 examples from the official document, one custom example
    assert_allclose(92.4296472453428, viscosity_index(73.3E-6, 8.86E-6))
    assert_allclose(156.42348293257797, viscosity_index(22.83E-6, 5.05E-6))
    assert_allclose(111.30701701381422, viscosity_index(53.47E-6, 7.80E-6))
    assert_allclose(92.03329369797858, viscosity_index(73.5E-6, 8.86E-6))
    assert_allclose(192.9975428057893, viscosity_index(1000E-6, 100E-6)) # Custom
    assert 193 == viscosity_index(1000E-6, 100E-6, rounding=True) # custom, rounded
    assert 92 == viscosity_index(73.3E-6, 8.86E-6, rounding=True)
    assert None == viscosity_index(3E-6, 1.5E-6)


@pytest.mark.meta_T_dept
def test_ViscosityLiquid():
    EtOH = ViscosityLiquid(MW=46.06844, Tm=159.05, Tc=514.0, Pc=6137000.0, Vc=0.000168, omega=0.635, Psat=7872.16, Vml=5.8676e-5, CASRN='64-17-5')

    mul_calcs = [(EtOH.set_user_methods(i), EtOH.T_dependent_property(298.15))[1] for i in EtOH.all_methods]
    mul_exp = [0.0010623746999654108, 0.0004191198228004424, 0.0010823506202025659, 0.0010720812586059742, 0.0010713697500000004, 0.0031157679801337825, 0.0010774308462863267, 0.0010823506202025659]
    assert_allclose(sorted(mul_calcs), sorted(mul_exp))

    # Test that methods return None
    EtOH.tabular_extrapolation_permitted = False
    Vml_calcs = [(EtOH.set_user_methods(i, forced=True), EtOH.T_dependent_property(600))[1] for i in EtOH.all_methods]
    assert [None]*8 == Vml_calcs

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')

    # Acetic acid to test ViswanathNatarajan2Exponential
    acetic_acid = ViscosityLiquid(CASRN='64-19-7', Tc=590.7)
    mul_calcs = [(acetic_acid.set_user_methods(i), acetic_acid.T_dependent_property(350.0))[1] for i in acetic_acid.all_methods]
    mul_exp = [0.0005744169247310638, 0.0005089289428076254, 0.0005799665143154318, 0.0005727888422607339, 0.000587027903931889]
    assert_allclose(sorted(mul_calcs), sorted(mul_exp))
    assert [None]*5 == [(acetic_acid.set_user_methods(i), acetic_acid.T_dependent_property(650.0))[1] for i in acetic_acid.all_methods]

    # Test ViswanathNatarajan2 with boron trichloride
    mu = ViscosityLiquid(CASRN='10294-34-5').T_dependent_property(250)
    assert_allclose(mu, 0.0003389255178814321)
    assert None == ViscosityLiquid(CASRN='10294-34-5').T_dependent_property(350)


    # Ethanol compressed
    EtOH = ViscosityLiquid(MW=46.06844, Tm=159.05, Tc=514.0, Pc=6137000.0, Vc=0.000168, omega=0.635, Psat=7872.16, Vml=5.8676e-5, CASRN='64-17-5')

    assert [False, True] == [EtOH.test_method_validity_P(300, P, COOLPROP) for P in (1E3, 1E5)]
    assert [True, True] == [EtOH.test_method_validity_P(300, P, LUCAS) for P in (1E3, 1E5)]

    assert_allclose(EtOH.calculate_P(298.15, 1E6, LUCAS), 0.0010880229239312313)
    assert_allclose(EtOH.calculate_P(298.15, 1E6, COOLPROP), 0.0010885493279015608)

    EtOH = ViscosityLiquid(MW=46.06844, Tm=159.05, Tc=514.0, Pc=6137000.0, Vc=0.000168, omega=0.635, Psat=7872.16, Vml=5.8676e-5, CASRN='64-17-5')
    # Ethanol data, calculated from CoolProp
    Ts = [275, 300, 350]
    Ps = [1E5, 5E5, 1E6]
    TP_data = [[0.0017455993713216815, 0.0010445175985089377, 0.00045053170256051774], [0.0017495149679815605, 0.0010472128172002075, 0.000452108003076486], [0.0017543973013034444, 0.0010505716944451827, 0.00045406921275411145]]
    EtOH.set_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_allclose(TP_data, recalc_pts)

    EtOH.tabular_extrapolation_permitted = False
    EtOH.forced_P = True
    assert None == EtOH.TP_dependent_property(300, 9E4)
    EtOH.tabular_extrapolation_permitted = True
    assert_allclose(EtOH.TP_dependent_property(300, 9E4), 0.0010445175985089377)

    with pytest.raises(Exception):
        EtOH.test_method_validity_P(300, 1E5, 'BADMETHOD')

@pytest.mark.meta_T_dept
def test_ViscosityGas():
    EtOH = ViscosityGas(MW=46.06844, Tc=514.0, Pc=6137000.0, Zc=0.2412, dipole=1.44, Vmg=0.02357, CASRN='64-17-5')

    mug_calcs = [(EtOH.set_user_methods(i), EtOH.T_dependent_property(298.15))[1] for i in EtOH.all_methods]
    mug_exp = [8.934627758386856e-06, 8.933684639927153e-06, 7.414017252400231e-06, 8.772549629893446e-06, 8.5445e-06, 7.902892297231681e-06, 8.805532218477024e-06, 7.536618820670175e-06]
    assert_allclose(sorted(mug_calcs), sorted(mug_exp))

    # Test that methods return None
    EtOH.tabular_extrapolation_permitted = False
    mug_calcs = [(EtOH.set_user_methods(i, forced=True), EtOH.T_dependent_property(6000))[1] for i in EtOH.all_methods]
    assert [None]*8 == mug_calcs

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')


    # Ethanol compressed
    EtOH = ViscosityGas(MW=46.06844, Tc=514.0, Pc=6137000.0, Zc=0.2412, dipole=1.44, Vmg=0.02357, CASRN='64-17-5')

    assert [True, False] == [EtOH.test_method_validity_P(300, P, COOLPROP) for P in (1E3, 1E5)]

    assert_allclose(EtOH.calculate_P(298.15, 1E3, COOLPROP), 8.77706377246337e-06)


    # Ethanol data, calculated from CoolProp
    Ts = [400, 500, 550]
    Ps = [1E3, 1E4, 1E5]
    TP_data = [[1.18634700291489e-05, 1.4762189560203758e-05, 1.6162732753470533e-05], [1.1862505513959454e-05, 1.4762728590964208e-05, 1.6163602669178767e-05], [1.1853229260926176e-05, 1.4768417536555742e-05, 1.617257402798515e-05]]
    EtOH.set_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_allclose(TP_data, recalc_pts)

    EtOH.tabular_extrapolation_permitted = False
    EtOH.forced_P = True
    assert None == EtOH.TP_dependent_property(300, 9E4)
    EtOH.tabular_extrapolation_permitted = True
    assert_allclose(EtOH.TP_dependent_property(300, 9E4), 1.1854259955707653e-05)

    with pytest.raises(Exception):
        EtOH.test_method_validity_P(300, 1E5, 'BADMETHOD')



def test_ViscosityLiquidMixture():
    # DIPPR  1983 manual example
    m = Mixture(['carbon tetrachloride', 'isopropanol'], zs=[0.5, 0.5], T=313.2)
    
    ViscosityLiquids = [i.ViscosityLiquid for i in m.Chemicals]

    obj = ViscosityLiquidMixture(ViscosityLiquids=ViscosityLiquids, CASs=m.CASs)
    mu = obj.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(mu, 0.0009956952502281852)
    
    mu = obj.calculate(m.T, m.P, m.zs, m.ws, MIXING_LOG_MOLAR)
    assert_allclose(mu, 0.0009956952502281852)
    mu = obj.calculate(m.T, m.P, m.zs, m.ws, MIXING_LOG_MASS)
    assert_allclose(mu, 0.0008741268796817256)
    
    # Test Laliberte
    m = Mixture(['water', 'sulfuric acid'], zs=[0.5, 0.5], T=298.15)
    ViscosityLiquids = [i.ViscosityLiquid for i in m.Chemicals]
    obj = ViscosityLiquidMixture(ViscosityLiquids=ViscosityLiquids, CASs=m.CASs)
    mu = obj.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(mu, 0.024955325569420893)
    assert obj.sorted_valid_methods == [LALIBERTE_MU]
    
    # Unhappy paths
    with pytest.raises(Exception):
        obj.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')
        
    with pytest.raises(Exception):
        obj.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')


def test_ViscosityGasMixture():
    # DIPPR  1983 manual example
    m = Mixture(['dimethyl ether', 'sulfur dioxide'], zs=[.95, .05], T=308.2)
    ViscosityGases = [i.ViscosityGas for i in m.Chemicals]
    obj = ViscosityGasMixture(MWs=m.MWs, molecular_diameters=m.molecular_diameters, Stockmayers=m.Stockmayers, CASs=m.CASs, ViscosityGases=ViscosityGases)
    
    mu =  obj.mixture_property(m.T, m.P, m.zs, m.ws)
    assert_allclose(mu, 9.637173494726528e-06)
    
    viscosity_gas_mixture_methods = [BROKAW, HERNING_ZIPPERER, WILKE, SIMPLE]
    mus = [obj.calculate(m.T, m.P, m.zs, m.ws, method) for method in viscosity_gas_mixture_methods]
    assert_allclose(mus, [9.637173494726528e-06, 9.672122280295219e-06, 9.642294904686337e-06, 9.638962759382555e-06])

    # Unhappy paths
    with pytest.raises(Exception):
        obj.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')
        
    with pytest.raises(Exception):
        obj.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')
